import dbus
from dbus.mainloop.glib import DBusGMainLoop
from gi.repository import GLib

from grapejuice_common.ipc.dbus_config import bus_path
from grapejuiced.dbus_service import DBusService


class State:
    def __init__(self, **kwargs):
        DBusGMainLoop(set_as_default=True)

        if "bus" in kwargs.keys():
            self.session_bus = kwargs["bus"]
        else:
            self.session_bus = dbus.SessionBus()

        if "start_service" in kwargs.keys() and kwargs["start_service"]:
            self.start_service()
        else:
            self.service = None

        self.loop = GLib.MainLoop()

    def start(self):
        self.loop.run()
        return self

    def stop(self):
        self.loop.quit()

    def start_service(self, bus=None):
        if bus is None:
            bus = self.session_bus

        if self.service is None:
            self.service = DBusService(bus, bus_path)



import logging

from grapejuice_common.ipc.i_dbus_connection import IDBusConnection

LOG = logging.getLogger(__name__)


class NoDaemonModeConnection(IDBusConnection):
    @property
    def connected(self):
        return True

    def launch_studio(self):
        from grapejuice_common import robloxctrl

        robloxctrl.run_studio()

    def play_game(self, uri):
        from grapejuice_common import robloxctrl

        def do_run():
            robloxctrl.run_player(uri)

        if robloxctrl.locate_player_launcher():
            do_run()

        else:
            robloxctrl.run_installer(post_install_function=do_run)

    def edit_local_game(self, place_path):
        from grapejuice_common import robloxctrl

        robloxctrl.run_studio(place_path, True)

    def edit_cloud_game(self, uri):
        from grapejuice_common import robloxctrl

        robloxctrl.run_studio(uri)

    def install_roblox(self):
        from grapejuice_common import robloxctrl

        robloxctrl.run_installer()

    def version(self):
        from grapejuiced import __version__

        return __version__

    def extract_fast_flags(self):
        from grapejuice_common import robloxctrl

        robloxctrl.fast_flag_extract()



import argparse
import logging
import random
import sys
from typing import Callable

import grapejuice_common.util
from grapejuice_common.gtk.gtk_stuff import gtk_boot
from grapejuice_common.ipc.dbus_client import dbus_connection
from grapejuice_common.logs import log_config
from grapejuice_common.logs.log_vacuum import vacuum_logs


def main_gui():
    def make_main_window():
        from grapejuice_common.logs import self_test
        self_test.post.run()

        from grapejuice.gui.main_window import MainWindow
        main_window = MainWindow()
        main_window.show()

    gtk_boot(make_main_window)


def func_gui(args):
    main_gui()


def func_player(args):
    def player_main():
        dbus_connection().play_game(grapejuice_common.util.prepare_uri(args.uri))

    gtk_boot(player_main, gtk_main=False)

    return 0


def func_app(*_):
    def player_main():
        import grapejuice_common.variables as v
        dbus_connection().play_game(grapejuice_common.util.prepare_uri(v.roblox_app_experience_url()))

    gtk_boot(player_main, gtk_main=False)

    return 0


def func_studio(args):
    uri = grapejuice_common.util.prepare_uri(args.uri)
    if uri:
        is_local = False
        if not uri.startswith("roblox-studio:"):
            uri = "Z:" + uri.replace("/", "\\")
            is_local = True

        if is_local:
            dbus_connection().edit_local_game(uri)
        else:
            dbus_connection().edit_cloud_game(uri)

    else:
        dbus_connection().launch_studio()


def func_install_roblox(args):
    from grapejuice_common import robloxctrl

    do_install = not (args.only_once and robloxctrl.locate_player_launcher())

    if do_install:
        robloxctrl.run_installer()


def func_uninstall_grapejuice(*_):
    from grapejuice_common import uninstall

    uninstall_grapejuice_response = input(
        "Are you sure you want to uninstall grapejuice? [y/N] "
    ).strip().lower()

    uninstall_grapejuice = (uninstall_grapejuice_response[0] == "y") if uninstall_grapejuice_response else False

    if uninstall_grapejuice:
        delete_prefix_response = input(
            "Remove the Wineprefix that contains your installation of Roblox Studio? This will cause all "
            "configuration of Roblox Studio to be permanently deleted! [n/Y] "
        ).strip().lower()

        delete_prefix = (delete_prefix_response[0] == "y") if delete_prefix_response else False

        params = uninstall.UninstallationParameters(delete_prefix, for_reals=True)
        uninstall.go(params)

        print("Grapejuice has been uninstalled, have a nice day!")

    else:
        print("Uninstallation aborted")


def run_daemon_instead(argv):
    from grapejuiced.__main__ import main as daemon_main
    daemon_main([sys.argv[0], *argv])

    return 0


def main(in_args=None):
    log_config.configure_logging("grapejuice")
    log = logging.getLogger(f"{__name__}/main")

    from grapejuice_common.features.settings import current_settings
    from grapejuice_common.update_info_providers import guess_relevant_provider

    update_info_provider = guess_relevant_provider()

    if current_settings:
        log.info("Loaded settings")

    if in_args is None:
        in_args = sys.argv

    if len(in_args) > 1:
        if in_args[1].lower() == "grapejuiced":
            return run_daemon_instead(in_args[2:])

    if random.randint(0, 10) == 5:
        print("beep beep")

    parser = argparse.ArgumentParser(prog="grapejuice", description="Manage Roblox on Linux")
    subparsers = parser.add_subparsers(title="subcommands", help="sub-command help")

    parser_gui = subparsers.add_parser("gui")
    parser_gui.set_defaults(func=func_gui)

    parser_player = subparsers.add_parser("player")
    parser_player.add_argument("uri", type=str, help="Your Roblox token to join a game")
    parser_player.set_defaults(func=func_player)

    parser_studio = subparsers.add_parser("studio")
    parser_studio.add_argument(
        "uri",
        nargs="?",
        type=str,
        help="The URI or file to open roblox studio with",
        default=None
    )

    parser_studio.set_defaults(func=func_studio)

    parser_install_roblox = subparsers.add_parser("install-roblox")
    parser_install_roblox.add_argument("--only-once", action="store_true")
    parser_install_roblox.set_defaults(func=func_install_roblox)

    if update_info_provider.can_update():
        parser_uninstall_grapejuice = subparsers.add_parser("uninstall")
        parser_uninstall_grapejuice.set_defaults(func=func_uninstall_grapejuice)

    parser_app = subparsers.add_parser("app")
    parser_app.set_defaults(func=func_app)

    args = parser.parse_args(in_args[1:])

    exit_code = 1

    if hasattr(args, "func"):
        f: Callable[[any], int] = getattr(args, "func")
        exit_code = f(args) or 0

    else:
        parser.print_help()

    try:
        log.info("Vacuuming logs")
        vacuum_logs()

    except Exception as e:
        # Vacuuming logs appears to break on some systems
        # So let's just catch any exception
        log.error(str(e))

    return exit_code


if __name__ == "__main__":
    sys.exit(main())



import json
import logging
from pathlib import Path
from typing import Dict

from grapejuice_common import variables

LOG = logging.getLogger(__name__)

k_show_fast_flag_warning = "show_fast_flag_warning"
k_wine_binary = "wine_binary"
k_dll_overrides = "dll_overrides"
k_no_daemon_mode = "no_daemon_mode"
k_release_channel = "release_channel"
k_environment_variables = "env"


def default_settings() -> Dict[str, any]:
    return {
        k_show_fast_flag_warning: True,
        k_wine_binary: "",
        k_dll_overrides: "ucrtbase=n,b;api-ms-win-crt-private-l1-1-0=n,b;dxdiagn=;winemenubuilder.exe=",
        k_no_daemon_mode: True,
        k_release_channel: "master",
        k_environment_variables: dict()
    }


class UserSettings:
    _settings_object: Dict[str, any] = None
    _location: Path = None

    def __init__(self, file_location=variables.grapejuice_user_settings()):
        self._location = file_location
        self.load()

    def get(self, key: str, default_value: any = None):
        if self._settings_object:
            return self._settings_object.get(key, default_value)

        return default_value

    def set(self, key: str, value: any, save: bool = False) -> any:
        self._settings_object[key] = value

        if save:
            self.save()

        return value

    def load(self):
        if self._location.exists():
            LOG.debug(f"Loading settings from '{self._location}'")

            try:
                save_settings = False

                with self._location.open("r") as fp:
                    self._settings_object = json.load(fp)

                    for k, v in default_settings().items():
                        if k not in self._settings_object:
                            self._settings_object[k] = v
                            save_settings = True

                if save_settings:
                    self.save()

            except json.JSONDecodeError as e:
                LOG.error(e)
                self._settings_object = default_settings()
                self.save()

        else:
            LOG.debug("There is no settings file present, going to save one")

            self.save()

    def save(self):
        LOG.debug(f"Saving settings file to '{self._location}'")

        with self._location.open("w+") as fp:
            self._settings_object = {
                **default_settings(),
                **(self._settings_object or {})
            }

            json.dump(self._settings_object, fp, indent=2)


current_settings = UserSettings()



import logging
import os
import time
from pathlib import Path
from typing import Union, List

from grapejuice_common import variables
from grapejuice_common import winectrl
from grapejuice_common.logs.log_util import log_function
from grapejuice_common.util import download_file

LOG = logging.getLogger(__name__)

DOWNLOAD_URL = "https://www.roblox.com/download/client"


def get_installer():
    install_path = variables.installer_path()

    if install_path.exists():
        LOG.debug(f"Removing old installer at {install_path}")
        os.remove(install_path)

    download_file(DOWNLOAD_URL, install_path)


def run_installer(post_install_function: callable = None):
    winectrl.create_prefix()
    get_installer()

    p = variables.installer_path()
    LOG.info(f"Running installer at {p}")

    winectrl.run_exe(p, post_run_function=post_install_function)


@log_function
def locate_in_versions(exe_name) -> Union[Path, None]:
    search_roots: List[Path] = [
        variables.wine_roblox_prog(),
        variables.wine_roblox_local_settings(),
        variables.wine_roblox_appdata_local()
    ]

    for root in search_roots:
        versions = root / "Versions"

        if root.exists() and versions.exists() and versions.is_dir():
            executable_path = versions / exe_name

            if executable_path.exists() and executable_path.is_file():
                return executable_path

            for version in Path(versions).glob("*"):
                if version.is_dir():
                    executable_path = version / exe_name

                    if executable_path.exists() and executable_path.is_file():
                        return executable_path

    return None


@log_function
def locate_roblox_exe(exe_name) -> Path:
    versioned = locate_in_versions(exe_name)

    if not versioned:
        location = variables.wine_roblox_prog() / "Versions" / exe_name

        if location.exists():
            return location

    return versioned


@log_function
def locate_studio_launcher() -> Path:
    return locate_roblox_exe("RobloxStudioLauncherBeta.exe")


@log_function
def locate_studio_exe() -> Path:
    return locate_in_versions("RobloxStudioBeta.exe")


@log_function
def locate_player_launcher() -> Path:
    return locate_in_versions("RobloxPlayerLauncher.exe")


@log_function
def locate_studio_app_settings() -> Union[Path, None]:
    studio_exe = locate_studio_exe()

    if studio_exe is None:
        return None

    return studio_exe.parent / "ClientSettings" / "ClientAppSettings.json"


@log_function
def locate_player_app_settings() -> Union[Path, None]:
    player_exe = locate_player_launcher()

    if player_exe is None:
        return None

    return player_exe.parent / "ClientSettings" / "ClientAppSettings.json"


def run_studio(uri="", ide=False):
    launcher = locate_studio_launcher()

    if launcher is None:
        return False

    if ide:
        winectrl.run_exe(launcher, "-ide", uri)

    else:
        if uri:
            winectrl.run_exe(launcher, uri)

        else:
            winectrl.run_exe(launcher, "-ide")

    return True


def studio_with_events(**events):
    studio_exe = locate_studio_exe()
    if studio_exe is None:
        return False

    args = [studio_exe]

    for k, v in events.items():
        args.append("-" + k)
        args.append(v)

    return winectrl.run_exe(*args, run_async=True)


def fast_flag_extract():
    fast_flag_path = variables.wine_roblox_studio_app_settings()

    if fast_flag_path.exists():
        os.remove(fast_flag_path)

    process = studio_with_events(startEvent="FFlagExtract", showEvent="NoSplashScreen")

    while True:
        if fast_flag_path.exists():
            stat = os.stat(fast_flag_path)

            if stat.st_size > 0:
                break

        time.sleep(0.5)

    if process:
        process.kill()


def run_player(uri):
    LOG.info("Running Roblox Player")
    player = locate_player_launcher()

    if player is None:
        LOG.warning("Could not locate player")
        return False

    LOG.info(f"Located player at {player}")

    winectrl.run_exe(player, uri)

    return True



import os
from pathlib import Path


def prepare_uri(uri):
    if uri is None:
        return None

    if os.path.exists(uri):
        return uri

    prepared_uri = uri.replace("'", "")
    if prepared_uri:
        return prepared_uri
    else:
        return None


def download_file(url, target_path: Path):
    import requests

    response = requests.get(url)
    response.raise_for_status()

    with open(target_path, "wb+") as fp:
        fp.write(response.content)

    return target_path



import os
import shutil
import subprocess
from datetime import datetime
from pathlib import Path

import grapejuice.__about__ as about
from grapejuice_common.util.task_sequence import TaskSequence
from grapejuice_packaging.builders.linux_package_builder import LinuxPackageBuilder
from grapejuice_packaging.util.distribution_detect import is_debian

ARCHITECTURE = "amd64"
STANDARDS_VERSION = "3.9.6"
DEBHELPER_COMPAT = 10
VERSION = f"{about.package_version}"
PACKAGE_VERSION = f"{VERSION}_{ARCHITECTURE}"
PREFIX = "/usr"
USR_LIB = os.path.join(PREFIX, "lib")
PYTHON_DIST_PACKAGES_DIR = f"{USR_LIB}/python3/dist-packages"
MAINTAINER = f"{about.author_name} <{about.author_email}>"
PACKAGE_FILENAME = f"{about.package_name}-{PACKAGE_VERSION}.deb"
DEBIAN_SECTION = "python"
DEBIAN_PRIORITY = "optional"
DEBIAN_DISTRIBUTION = "unstable"
DEBIAN_URGENCY = "medium"

FIELD_SOURCE = ("Source", about.package_name)
FIELD_MAINTAINER = ("Maintainer", MAINTAINER)
FIELD_STANDARDS_VERSION = ("Standards-Version", STANDARDS_VERSION)
FIELD_ARCHITECTURE = ("Architecture", ARCHITECTURE)
FIELD_BUILD_DEPENDS = ("Build-Depends", [
    "debhelper",
    "python3",
    "python3-pip",
    "python3-virtualenv",
    "git", "unzip"
])

CONTROL_FIELDS = [
    FIELD_SOURCE,
    ("Section", DEBIAN_SECTION),
    ("Priority", DEBIAN_PRIORITY),
    FIELD_MAINTAINER,
    FIELD_BUILD_DEPENDS,
    FIELD_STANDARDS_VERSION,
    None,
    ("Package", about.package_name),
    FIELD_ARCHITECTURE,
    ("Depends", [
        "python3 (>= 3.7~)",
        "python3-packaging",
        "python3-setuptools",
        "python3-psutil",
        "python3-requests",
        "python3-gi",
        "libcairo2",
        "libgirepository-1.0-1",
        "libgtk-3-0",
        "libgtk-3-bin",
        "gobject-introspection",
        "gir1.2-gtk-3.0"
    ]),
    ("Homepage", about.package_repository),
    ("Description", about.package_description + "\n Generated by grapejuice_packaging")
]

COPYRIGHT_FIELDS = [
    ("Format", "https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/"),
    ("Upstream-Name", about.project_name),
    ("Upstream-Contact", MAINTAINER),
    FIELD_SOURCE,
    None,
    ("Files", "*"),
    ("Copyright", f"2019-{datetime.now().year} {about.author_name}"),
    ("License", about.package_license)
]

RULES = """#!/usr/bin/make -f

%:
\tdh $@

override_dh_shlibdeps:
\ttrue
"""


def _fields_to_string(fields):
    def process_field(field):
        if field is None:
            return ""

        assert isinstance(field, (list, tuple))
        assert len(field) == 2

        key, value = field

        if isinstance(value, list):
            value = ", ".join(value)

        return f"{key}: {str(value)}"

    return "\n".join(list(map(process_field, fields)))


class DebianPackageBuilder(LinuxPackageBuilder):
    _package_files_loc: str = None

    def build(self):
        self.clean_build()
        self._prepare_build()

        self._package_files_loc = self._build_dir
        self._build_dir = os.path.join(self._build_dir, "pkg")

        super_build = super().build
        build = TaskSequence("Build Debian Package")

        @build.task("Create the base linux package")
        def create_linux_package(log):
            build_root = self._build_dir

            self._build_dir = os.path.join(build_root, "ROOT")
            self._configuration.root = self._build_dir

            super_build()

            self._build_dir = build_root
            self._configuration.root = self._build_dir

        @build.task("Create debian directory")
        def create_debian_directory(log):
            path = Path(self._build_dir, "debian")
            os.makedirs(path, exist_ok=True)

        @build.task("Write compat file")
        def write_compat(log):
            path = Path(self._build_dir, "debian", "compat")

            with path.open("w+") as fp:
                fp.write(str(int(DEBHELPER_COMPAT)))

        @build.task("Write install file")
        def write_install_file(log):
            path = Path(self._build_dir, "debian", "install")

            with path.open("w+") as fp:
                content = "ROOT/* /\n"
                fp.write(content)

        @build.task("Write control file")
        def write_control(log):
            path = Path(self._build_dir, "debian", "control")

            with path.open("w+") as fp:
                fp.write(_fields_to_string(CONTROL_FIELDS))

        @build.task("Write copyright file")
        def write_copyright(log):
            path = Path(self._build_dir, "debian", "copyright")

            with path.open("w+") as fp:
                fp.write(_fields_to_string(COPYRIGHT_FIELDS))

        @build.task("Write files file")
        def write_files(log):
            path = Path(self._build_dir, "debian", "files")

            with path.open("w+") as fp:
                fp.write(" ".join([PACKAGE_FILENAME, DEBIAN_SECTION, DEBIAN_PRIORITY]))

        @build.task("Write rules file")
        def write_rules(log):
            path = Path(self._build_dir, "debian", "rules")

            with path.open("w+") as fp:
                fp.write(RULES)

        @build.task("Write changelog file")
        def write_changelog(log):
            lines = [f"{about.package_name} ({VERSION}) {DEBIAN_DISTRIBUTION}; urgency={DEBIAN_URGENCY}\n", "\n"]

            try:
                changelog = subprocess.check_output(["git", "shortlog"], timeout=1) \
                    .decode("UTF - 8").strip().split("\n")

            except subprocess.TimeoutExpired as e:
                changelog = ["Automatically generated package"]
                print(str(e))

            for changelog_line in list(map(lambda s: "  * " + s, changelog)):
                lines.append(changelog_line + "\n")

            lines.append("\n")

            date_str = datetime.now().strftime("%a, %d %b %Y %H:%M:%S") + " +0000"
            lines.append(f" -- {MAINTAINER}  {date_str}\n")

            with open(os.path.join(self._build_dir, "debian", "changelog"), "w+") as fp:
                fp.writelines(lines)

        build.run()

    def dist(self):
        self.clean_dist()
        self._prepare_dist()

        dist = TaskSequence("Create distribution files for debian")

        @dist.task("Strip out __pycache__ directories")
        def strip_py_cache(log):
            for directory in Path(self._build_dir).rglob("__pycache__"):
                if directory.is_dir():
                    log.info(f"Removing {directory}")
                    shutil.rmtree(directory, ignore_errors=True)

        if is_debian():
            wd = os.getcwd()

            @dist.task("Create package file")
            def create_package(log):
                os.chdir(self._build_dir)
                subprocess.check_call(["debuild", "-uc", "-us"])
                os.chdir(wd)

        if self._package_files_loc is not None:
            @dist.task("Move distribution files")
            def move_files(log):
                for file in Path(self._package_files_loc).glob("*"):
                    if file.is_file():
                        src = str(file.absolute())
                        dst = Path(self._dist_dir, file.name)

                        log.info(f"Moving file {src} -> {dst}")
                        shutil.move(src, dst)

        dist.run()



import os


def os_release():
    search_paths = [
        "/etc/os-release",
        "/usr/lib/os-release"
    ]

    for path in search_paths:
        if os.path.exists(path):
            with open(path, "r") as fp:
                return fp.read()

    raise RuntimeError("Could not open os-release")


def is_debian():
    return "ID=debian" in os_release()


def is_arch():
    return "ID=arch" in os_release()



import dbus.service

from grapejuice_common.ipc.dbus_config import bus_name
from grapejuiced.__init__ import __version__


class DBusService(dbus.service.Object):
    def __init__(self, bus, path):
        super().__init__(bus, path, dbus.service.BusName(bus_name))
        self.version_string = str(__version__)

    @dbus.service.method(
        dbus_interface=bus_name,
        in_signature="s",
        out_signature="b"
    )
    def EditLocalGame(self, path):
        from grapejuice_common import robloxctrl

        return robloxctrl.run_studio(path, True)

    @dbus.service.method(
        dbus_interface=bus_name,
        in_signature="s",
        out_signature="b"
    )
    def EditCloudGame(self, uri):
        from grapejuice_common import robloxctrl
        return robloxctrl.run_studio(uri)

    @dbus.service.method(
        dbus_interface=bus_name,
        in_signature="",
        out_signature="b"
    )
    def LaunchStudio(self):
        from grapejuice_common import robloxctrl
        return robloxctrl.run_studio()

    @dbus.service.method(
        dbus_interface=bus_name,
        in_signature="s",
        out_signature=""
    )
    def PlayGame(self, uri):
        from grapejuice_common import robloxctrl

        def do_run():
            robloxctrl.run_player(uri)

        if robloxctrl.locate_player_launcher():
            do_run()

        else:
            robloxctrl.run_installer(post_install_function=do_run)

    @dbus.service.method(
        dbus_interface=bus_name,
        in_signature="",
        out_signature=""
    )
    def InstallRoblox(self):
        from grapejuice_common import robloxctrl
        robloxctrl.run_installer()

    @dbus.service.method(
        dbus_interface=bus_name,
        in_signature="",
        out_signature=""
    )
    def ExtractFastFlags(self):
        from grapejuice_common import robloxctrl
        robloxctrl.fast_flag_extract()

    @dbus.service.method(
        dbus_interface=bus_name,
        in_signature="",
        out_signature="s"
    )
    def Version(self):
        return self.version_string



#!/usr/bin/env python3
import os
import subprocess
import sys

if os.getuid() == 0:
    if "CI_JOB_ID" not in os.environ:
        msg = "Installing Grapejuice as root is not supported"
        print(msg, file=sys.stderr)
        raise RuntimeError(msg)

try:
    architecture = subprocess.check_output(["uname", "-m"]).decode("UTF-8").strip()
    expected_architecture = "x86_64"

    if architecture.lower() != expected_architecture:
        msg = f"Roblox Studio will only run on the {expected_architecture} CPU architecture. " \
              f"The CPU architecture of this machine is {architecture}. The installer will now exit."
        print(msg, file=sys.stderr)
        sys.exit(-1)

except subprocess.CalledProcessError:
    pass

REQUIRED_MAJOR = 3
REQUIRED_MINOR = 7


def perform_install():
    from pathlib import Path

    subprocess.check_call([sys.executable, "setup.py", "install_locally"])

    if "CI_JOB_ID" not in os.environ:
        subprocess.check_call([
            str(Path(os.environ["HOME"], ".local", "bin", "grapejuice").resolve()),
            "install-roblox", "--only-once"
        ])


def have_tkinter():
    try:
        import tkinter
        return True
    except ImportError:
        return False


def err_tkinter(title, message):
    import tkinter
    from tkinter import messagebox

    root = tkinter.Tk()
    root.withdraw()

    messagebox.showerror(title, message)


def have_zenity():
    import os
    return os.path.exists("/usr/bin/zenity")


def err_zenity(title, message):
    subprocess.call(["zenity", "--error", title, "--no-wrap", "--text={}".format(message)])


def err_desperation(message):
    subprocess.call(["xmessage", message])


def show_err(title, message):
    if have_tkinter():
        err_tkinter(title, message)

    elif have_zenity():
        err_zenity(title, message)

    else:
        err_desperation(message)


def err_py37():
    import sys

    ver = f"{REQUIRED_MAJOR}.{REQUIRED_MINOR}"

    show_err("Out of date",
             f"Your current version of python is out of date and therefore Grapejuice cannot be installed.\n\n"
             f"Python {ver} is required. Check the Grapejuice source repository for the installation instructions.\n\n"
             f"You have:\n{sys.version}"
             )


def have_py37():
    import sys

    satisfied = sys.version_info.major >= REQUIRED_MAJOR and sys.version_info.minor >= REQUIRED_MINOR

    if not satisfied:
        exit_code = -1

        try:
            err_py37()

        except Exception as e:
            exit_code = -2
            print(e, file=sys.stderr)

        sys.exit(exit_code)

    return satisfied


if __name__ == "__main__":
    if have_py37():
        perform_install()



import os
import shutil
import subprocess
import sys
from pathlib import Path

import grapejuice_common.variables as v
from grapejuice_common.util.task_sequence import TaskSequence
from grapejuice_packaging import packaging_resources
from grapejuice_packaging.builders.package_builder import PackageBuilder


class AppImageBuilder(PackageBuilder):
    _script_name = "__main__"

    def build(self):
        assert os.path.exists("setup.py") and os.path.isdir("src")

        app_image = TaskSequence("Build AppImage")
        app_dir = os.path.join(self._build_dir, "grapejuice.AppDir")
        pyinstaller_output_path = os.path.join(self._dist_dir, self._script_name)

        @app_image.task("Clean workspace")
        def clean(log):
            self.clean_build()
            self.clean_dist()

        @app_image.task("Run PyInstaller")
        def run_pyinstaller(log):
            script_path = os.path.join(os.path.abspath(os.getcwd()), "src", "grapejuice", f"{self._script_name}.py")
            log.info(f"Running PyInstaller on: {script_path}")

            subprocess.check_call([
                sys.executable, "-m", "PyInstaller",
                script_path
            ])

        @app_image.task("Strip binaries")
        def strip_binaries(log):
            pyinstaller_root = Path(pyinstaller_output_path)
            log.info(f"PyInstaller root: {pyinstaller_root}")

            rubbish = []

            for file in pyinstaller_root.rglob("*"):
                if "libstdc" in file.name.lower():
                    rubbish.append(file)

            for file in rubbish:
                path = str(file.absolute())
                log.info(f"Removing file: {path}")

                os.remove(path)

        @app_image.task("Prepare AppDir")
        def prepare_app_dir(log):
            log.info(f"AppDir is: {app_dir}")
            os.makedirs(app_dir, exist_ok=True)

            app_src = pyinstaller_output_path
            app_dst = os.path.join(app_dir, "app")
            log.info(f"Copying tree: {app_src} -> {app_dst}")

            shutil.copytree(app_src, app_dst)

        @app_image.task("Copy assets")
        def copy_assets(log):
            assets_src = v.assets_dir()
            assets_dst = os.path.join(app_dir, "app", "assets")
            log.info(f"Copying tree: {assets_src} -> {assets_dst}")

            shutil.copytree(assets_src, assets_dst)

        @app_image.task("Copy AppRun")
        def copy_app_run(log):
            src = packaging_resources.app_image_app_run()
            dst = os.path.join(app_dir, "AppRun")
            log.info(f"Copying AppRun: {src} -> {dst}")

            shutil.copy(src, dst)
            os.chmod(dst, 0o755)

        @app_image.task("Copy desktop entry")
        def copy_desktop_entry(log):
            src = packaging_resources.app_image_desktop()
            dst = os.path.join(app_dir, "grapejuice.desktop")
            log.info(f"Copying desktop entry: {src} -> {dst}")

            shutil.copy(src, dst)

        @app_image.task("Copy application icon")
        def copy_application_info(log):
            src = v.grapejuice_icon()
            dst = os.path.join(app_dir, "grapejuice.svg")
            log.info(f"Copying icon: {src} -> {dst}")

            shutil.copy(src, dst)

        app_image.run()

    def dist(self):
        app_image = TaskSequence("Create Appimage distribution")
        app_dir = os.path.join(self._build_dir, "grapejuice.AppDir")
        app_dir_parent = os.path.dirname(app_dir)

        @app_image.task("Clean distribution folder again")
        def clean_dist(log):
            self.clean_dist()
            os.makedirs(self._dist_dir, exist_ok=True)

        @app_image.task("Run AppImage tool")
        def run_app_image_tool(log):
            log.info(f"Changing directory to: {app_dir_parent}")
            os.chdir(app_dir_parent)

            path = packaging_resources.app_image_tool()
            log.info(f"Using AppImage tool at: {path}")

            os.chmod(path, 0o755)
            subprocess.check_call([path, app_dir])

        @app_image.task("Move AppImage to dist folder")
        def move_app_image(log):
            for file in Path(app_dir_parent).rglob("*.AppImage"):
                shutil.move(str(file.absolute()), os.path.join(self._dist_dir, file.name))

        app_image.run()



import sys


def gtk_boot(main_function, *args, gtk_main=True, **kwargs):
    assert callable(main_function)
    sys.argv[0] = "Grapejuice"

    import gi
    gi.require_version("Gtk", "3.0")
    from gi.repository import Gtk

    main_function(*args, **kwargs)

    if gtk_main:
        Gtk.main()


def dialog(dialog_text):
    from gi.repository import Gtk

    gtk_dialog = Gtk.MessageDialog(
        message_type=Gtk.MessageType.INFO,
        buttons=Gtk.ButtonsType.OK,
        text=dialog_text
    )

    gtk_dialog.run()
    gtk_dialog.destroy()


class WindowBase:
    def __init__(self, glade_path, handlers=None):
        self._glade_path = glade_path
        self._handlers = handlers

        self._build()

    def _create_builder(self):
        from gi.repository import Gtk

        builder = Gtk.Builder()
        builder.add_from_file(self._glade_path)

        return builder

    def _build(self):
        self.builder = self._create_builder()

        if self._handlers:
            self.builder.connect_signals(self._handlers)



import os
import subprocess
import sys
import tarfile
from pathlib import Path

from setuptools import Command

import grapejuice_common.variables as v
from grapejuice_common.util.task_sequence import TaskSequence

PYTHON_INTERPRETER = sys.executable

ROBLOX_STUDIO = "roblox-studio.desktop"
ROBLOX_PLAYER = "roblox-player.desktop"

MIME = {
    "x-scheme-handler/roblox-studio": ROBLOX_STUDIO,
    "x-scheme-handler/roblox-player": ROBLOX_PLAYER,
    "application/x-roblox-rbxl": ROBLOX_STUDIO,
    "application/x-roblox-rbxlx": ROBLOX_STUDIO
}


def _xdg_mime_default(desktop_entry: str, mime: str):
    subprocess.check_call(["xdg-mime", "default", desktop_entry, mime])


def _do_install(*_):
    assert os.path.exists("setup.py"), \
        "Project file not found, make sure you're in the Grapejuice root!"

    src_path = os.path.join(os.path.abspath(os.getcwd()), "src")
    if "PYTHONPATH" in os.environ:
        os.environ["PYTHONPATH"] = src_path + ":" + os.environ["PYTHONPATH"]

    else:
        os.environ["PYTHONPATH"] = src_path

    install = TaskSequence("Install Grapejuice locally")

    @install.task("Build package of supplemental files")
    def build_supplemental(log):
        subprocess.check_call([
            PYTHON_INTERPRETER, "-m", "grapejuice_packaging",
            "supplemental_package"
        ])

    @install.task("Install supplemental packages")
    def install_supplemental_packages(log):
        for file in Path("dist", "supplemental_package").glob("*.tar.gz"):
            log.info(f"Installing supplemental package {file}")

            with tarfile.open(file) as tar:
                tar.extractall(v.home())

    @install.task("Install Grapejuice package")
    def install_package(log):
        env_snapshot = None

        if "VIRTUAL_ENV" in os.environ:
            virtual_env = os.environ["VIRTUAL_ENV"]
            log.warning(f"Breaking out of virtualenv: {virtual_env}")
            env_snapshot = dict(os.environ)

            path = os.environ["PATH"].split(os.pathsep)
            path = list(filter(lambda s: not s.startswith(virtual_env), path))
            os.environ["PATH"] = os.pathsep.join(path)

            log.info("Set PATH to: " + os.environ["PATH"])
            os.environ.pop("VIRTUAL_ENV", None)

        subprocess.check_call([
            PYTHON_INTERPRETER, "-m", "pip",
            "install", ".",
            "--user",
            "--upgrade"
        ])

        if env_snapshot is not None:
            log.info("Restoring environment snapshot...")
            for env_key, env_value in env_snapshot.items():
                os.environ[env_key] = env_value

    @install.task("Updating GTK icon cache")
    def update_icon_cache(log):
        subprocess.check_call(["gtk-update-icon-cache"])

    @install.task("Updating desktop database")
    def update_desktop_database(log):
        path = Path(v.home(), ".local", "share", "applications").absolute()
        log.info(f"Updating desktop database: {path}")

        subprocess.check_call(["update-desktop-database", str(path)])

    @install.task("Updating MIME type associations")
    def update_mime_associations(log):
        for mime, desktop in MIME.items():
            log.info(f"Associating {mime} with {desktop}")
            _xdg_mime_default(desktop, mime)

    @install.task("Updating MIME database")
    def update_mime_database(log):
        path = Path(v.home(), ".local", "share", "mime").absolute()
        log.info(f"Updating MIME database: {path}")

        subprocess.check_call(["update-mime-database", str(path)])

    install.run()


class InstallLocally(Command):
    description = "Install Grapejuice locally"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        _do_install(self.user_options)



import atexit
import getpass
import logging
import os
import subprocess
import uuid
from pathlib import Path

from grapejuice_common.util.errors import NoWineError

HERE = os.path.abspath(os.path.dirname(__file__))
INSTANCE_ID = str(uuid.uuid4())

LOG = logging.getLogger(__name__)

at_exit_handlers = dict()


def ensure_dir(p):
    if not os.path.exists(p):
        os.makedirs(p)

    return p


def home():
    return str(Path(os.environ["HOME"]).resolve())


def system_application_dir():
    p = os.path.dirname(src_dir())

    assert os.path.exists(p)

    return p


def user_application_dir():
    return os.path.join(local_share(), "grapejuice")


def application_manifest():
    return os.path.join(user_application_dir(), "package_manifest.json")


def assets_dir():
    search_locations = [
        os.path.join(HERE, "assets"),
        os.path.join(os.getcwd(), "assets"),
        os.path.join(system_application_dir(), "assets")
    ]

    for p in search_locations:
        if os.path.exists(p):
            return p

    raise RuntimeError("Could not find assets directory")


def desktop_assets_dir():
    return os.path.join(assets_dir(), "desktop")


def mime_xml_assets_dir():
    return os.path.join(assets_dir(), "mime_xml")


def icons_assets_dir():
    return os.path.join(assets_dir(), "icons")


def grapejuice_icon():
    return os.path.join(icons_assets_dir(), "hicolor", "scalable", "apps", "grapejuice.svg")


def src_dir():
    return os.path.abspath(os.path.dirname(HERE))


def glade_dir():
    return os.path.join(assets_dir(), "glade")


def grapejuice_glade():
    return os.path.join(glade_dir(), "grapejuice.glade")


def about_glade():
    return os.path.join(glade_dir(), "about.glade")


def fast_flag_editor_glade():
    return os.path.join(glade_dir(), "fast_flag_editor.glade")


def grapejuice_components_glade():
    return os.path.join(glade_dir(), "grapejuice_components.glade")


def fast_flag_warning_glade():
    return os.path.join(glade_dir(), "fast_flag_warning.glade")


def config_base_dir():
    return ensure_dir(os.path.join(xdg_config_home(), "brinkervii"))


def grapejuice_config_dir():
    return ensure_dir(os.path.join(config_base_dir(), "grapejuice"))


def grapejuice_user_settings():
    return Path(grapejuice_config_dir(), "user_settings.json")


def wineprefix_dir():
    return os.path.join(user_application_dir(), "wineprefix")


def wine_drive_c():
    return os.path.join(wineprefix_dir(), "drive_c")


def wine_user_reg():
    return os.path.join(wineprefix_dir(), "user.reg")


def wine_roblox_prog() -> Path:
    return Path(wine_drive_c(), "Program Files (x86)", "Roblox")


def wine_roblox_local_settings() -> Path:
    return Path(wine_drive_c(), "users", getpass.getuser(), "Local Settings", "Application Data", "Roblox")


def wine_temp():
    return ensure_dir(os.path.join(wine_drive_c(), "windows", "temp"))


def wine_user():
    return os.path.join(wine_drive_c(), "users", os.environ["USER"])


def wine_roblox_appdata():
    p = os.path.join(wine_user(), "Local Settings", "Application Data", "Roblox")
    if os.path.exists(p):
        return p

    p = os.path.join(wine_user(), "AppData", "Local", "Roblox")
    return p


def wine_roblox_appdata_local() -> Path:
    return Path(wine_user(), "AppData", "Local", "Roblox")


def wine_roblox_global_settings_13():
    return os.path.join(wine_roblox_appdata(), "GlobalSettings_13.xml")


def wine_roblox_studio_app_settings():
    return Path(wine_roblox_appdata(), "ClientSettings", "StudioAppSettings.json")


def installer_path():
    return Path(wine_temp(), "Roblox_Installer.exe")


def xdg_config_home():
    if "XDG_CONFIG_HOME" in os.environ:
        config_home = os.environ["XDG_CONFIG_HOME"]
        if config_home and os.path.exists(config_home) and os.path.isdir(config_home):
            return config_home

    config_home = os.path.join(home(), ".config")
    if not os.path.exists(config_home):
        os.makedirs(config_home)

    return config_home


def dot_local():
    path = os.path.join(home(), ".local")
    if not os.path.exists(path):
        os.makedirs(path)

    return path


def local_share():
    return os.path.join(dot_local(), "share")


def local_var():
    return os.path.join(dot_local(), "var")


def local_log():
    return os.path.join(local_var(), "log")


def logging_directory():
    return os.path.join(local_log(), "grapejuice")


def xdg_documents():
    run = subprocess.run(["xdg-user-dir", "DOCUMENTS"], stdout=subprocess.PIPE, check=False)
    documents_path = run.stdout.decode("utf-8").rstrip()

    if os.path.exists(documents_path):
        return documents_path

    documents_path = os.path.join(home(), "Documents")
    return ensure_dir(documents_path)


def roblox_app_experience_url():
    return "roblox-player:+launchmode:app+robloxLocale:en_us+gameLocale:en_us+LaunchExp:InApp"


def roblox_return_to_studio():
    return "https://www.roblox.com/login/return-to-studio"


def git_repository():
    return "https://gitlab.com/brinkervii/grapejuice"


def git_wiki():
    return f"{git_repository()}/-/wikis/home"


def git_grapejuice_init():
    from grapejuice_common.features.settings import current_settings
    from grapejuice_common.features import settings

    release_channel = current_settings.get(settings.k_release_channel)
    return f"{git_repository()}/-/raw/{release_channel}/src/grapejuice/__init__.py"


def git_source_tarball():
    from grapejuice_common.features.settings import current_settings
    from grapejuice_common.features import settings

    release_channel = current_settings.get(settings.k_release_channel)
    return f"{git_repository()}/-/archive/{release_channel}/grapejuice-{release_channel}.tar.gz"


def tmp_path():
    path = Path(os.path.sep, "tmp", f"grapejuice-{INSTANCE_ID}").resolve()
    path_string = str(path)

    if "clean_tmp_path" not in at_exit_handlers:
        def on_exit(*_, **__):
            import shutil
            shutil.rmtree(path, ignore_errors=True)

        at_exit_handlers["clean_tmp_path"] = on_exit

    return ensure_dir(path_string)


def wine_binary(arch=""):
    from grapejuice_common.features.settings import current_settings
    from grapejuice_common.features import settings

    custom_wine_binary = (current_settings.get(settings.k_wine_binary) or "").strip()
    if custom_wine_binary and not os.path.exists(custom_wine_binary):
        LOG.error(f"!! Custom wine binary set to {custom_wine_binary}, but it does not exist !!")

    path_search = [custom_wine_binary] if custom_wine_binary else []

    if "PATH" in os.environ:
        for spec in os.environ["PATH"].split(":"):
            path_search.append(os.path.join(spec, "wine" + arch))

    static_search = [
        "/opt/wine-stable/bin/wine" + arch,
        "/opt/wine-staging/bin/wine" + arch
    ]

    for p in path_search + static_search:
        if p and os.path.exists(p):
            LOG.debug(f"Using wine binary at: {p}")
            return p

    raise NoWineError()


def wine_binary_64():
    return wine_binary("64")


def required_wine_version():
    return "wine-4.0"


def at_exit_handler(*args, **kwargs):
    for v in filter(callable, at_exit_handlers.values()):
        v(*args, **kwargs)


atexit.register(at_exit_handler)



import logging
import os

# Note about the self-test:
# Since part of the self-test is importing all project dependencies, Gtk gets imported.
# Gtk will cache sys.argv[0] and use this to display in the activities window.
# SO VERY IMPORTANT: MAKE SURE sys.arg[0] is hacked BEFORE Gtk is imported!
from grapejuice_common.util.task_sequence import TaskSequence, TaskError

post = TaskSequence("Power On Self Test")


@post.task("Read OS information")
def read_os_release(log: logging.Logger):
    os_release_path = "/etc/os-release"

    if not os.path.exists(os_release_path):
        raise TaskError(f"OS release information does not exist at '{os_release_path}'", can_continue=True)

    log.info(f"-- Path is {os_release_path}")
    with open(os_release_path, "r") as fp:
        for line in fp.readlines():
            log.info(line.strip())


@post.task("Test for root user")
def test_for_root_user(log: logging.Logger):
    if os.getuid() == 0:
        if "CI_JOB_ID" not in os.environ:
            msg = "Running Grapejuice as root is not supported"
            log.error(msg)
            raise TaskError(msg, can_continue=False)


@post.task("Check python dependencies")
def check_python_dependencies(log: logging.Logger):
    def import_psutil():
        import psutil

        return psutil

    def import_py_g_object():
        import gi
        gi.require_version("Gtk", "3.0")

        from gi.repository import Gtk
        from gi.repository import GObject
        from gi.repository import GLib

        return gi, Gtk, GObject, GLib

    def import_packaging():
        import packaging

        return packaging

    def import_requests():
        import requests

        return requests

    def import_setuptools():
        import setuptools

        return setuptools

    def try_import(import_func: callable):
        try:
            import_result = func()
            if isinstance(import_result, tuple):
                results = list(import_result)

            else:
                results = [import_result]

            log.info(f"Import function '{import_func.__name__}' loaded {len(results)} imports")

        except ImportError as e:
            msg = f"Import function '{import_func.__name__}' could not perform the imports it had to. " \
                  f"This installation of grapejuice is incorrect and should be rectified by doing a proper " \
                  f"installation. The underlying error is:\n {str(e)}"

            raise TaskError(msg, can_continue=False) from e

    dependency_check_functions = [
        import_psutil,
        import_py_g_object,
        import_packaging,
        import_requests,
        import_setuptools
    ]

    for func in dependency_check_functions:
        try_import(func)



import logging
import os
from typing import Iterable

from grapejuice import background
from grapejuice.gui.yes_no_dialog import yes_no_dialog
from grapejuice.tasks import DisableMimeAssociations, InstallRoblox, SandboxWine, \
    RunRobloxStudio, ExtractFastFlags, OpenLogsDirectory, PerformUpdate
from grapejuice_common import variables, robloxctrl, uninstall, update_info_providers
from grapejuice_common import winectrl
from grapejuice_common.features import settings
from grapejuice_common.features.settings import current_settings
from grapejuice_common.gtk.gtk_stuff import WindowBase, dialog
from grapejuice_common.registry_utils import logged_into_studio
from grapejuice_common.update_info_providers import UpdateInformationProvider
from grapejuice_common.util.errors import NoWineError
from grapejuice_common.util.event import Event

LOG = logging.getLogger(__name__)

on_destroy = Event()

once_task_tracker = dict()
on_background_task_error = Event()
on_background_errors_shown = Event()
background_task_errors = []

update_provider: UpdateInformationProvider = update_info_providers.guess_relevant_provider()


def on_task_removed(task: background.BackgroundTask):
    if task in once_task_tracker.keys():
        once_task_tracker[task] = None


background.tasks.task_removed.add_listener(on_task_removed)


def run_task_once(task_class, on_already_running: callable, *args, **kwargs):
    if task_class in once_task_tracker.values():
        on_already_running()
        return

    def on_error(*args2):
        on_background_task_error(*args2)

    kwargs["on_error_callback"] = on_error

    task = task_class(*args, **kwargs)
    once_task_tracker[task] = task_class

    background.tasks.add(task)


def generic_already_running():
    dialog("This task is already being performed!")


def xdg_open(*args):
    os.spawnlp(os.P_NOWAIT, "xdg-open", "xdg-open", *args)


class MainWindowHandlers:
    _updating = False

    def on_destroy(self, *_):
        from gi.repository import Gtk
        on_destroy()
        Gtk.main_quit()

    def run_winecfg(self, *_):
        winectrl.winecfg()

    def run_regedit(self, *_):
        winectrl.regedit()

    def run_winetricks(self, *_):
        winectrl.wine_tricks()

    def disable_mime_assoc(self, *_):
        run_task_once(DisableMimeAssociations, generic_already_running)

    def sandbox(self, *_):
        run_task_once(SandboxWine, generic_already_running)

    def run_roblox_installer(self, *_):
        def no_wine_dialog() -> None:
            dialog("Grapejuice could not find a working Wine binary, please install Wine using your operating "
                   "system's package manager in order to install and use Roblox.")

        try:
            wine_bin = variables.wine_binary()
            if not os.path.exists(wine_bin):
                no_wine_dialog()
                return

        except NoWineError:
            no_wine_dialog()
            return

        run_task_once(InstallRoblox, generic_already_running)

    def run_roblox_studio(self, *_):
        studio_launcher_location = robloxctrl.locate_studio_launcher()
        if not studio_launcher_location:
            dialog("Grapejuice could not locate Roblox Studio. You might have to install it first by going to the "
                   "maintenance tab and clicking 'Install Roblox'")
            return

        if not logged_into_studio() and yes_no_dialog(
            "Log into Roblox",
            "You are currently not signed into Roblox Studio. "
            "Roblox Studio is known to require an account to use. Would you like to sign in now?"
        ):
            xdg_open(variables.roblox_return_to_studio())

        else:
            run_task_once(RunRobloxStudio, generic_already_running)

    def wine_explorer(self, *_):
        winectrl.explorer()

    def open_drive_c(self, *_):
        xdg_open(variables.wine_drive_c())

    def show_about(self, *_):
        from grapejuice.gui.about_window import AboutWindow
        wnd = AboutWindow()
        wnd.window.run()

        del wnd

    def open_roblox_return_to_studio(self, *_):
        xdg_open(variables.roblox_return_to_studio())

    def open_fast_flag_editor(self, *_):
        def open_editor(b):
            if not b:
                return

            task = ExtractFastFlags()

            def poll():
                if task.finished:
                    from grapejuice.gui.fast_flag_editor import FastFlagEditor
                    wnd = FastFlagEditor()
                    wnd.window.show()

                return not task.finished

            from gi.repository import GObject
            GObject.timeout_add(100, poll)

            background.tasks.add(task)

        if current_settings.get(settings.k_show_fast_flag_warning):
            from grapejuice.gui.fast_flag_warning import FastFlagWarning
            wnd = FastFlagWarning(open_editor)
            wnd.show()

        else:
            open_editor(True)

    def show_wiki(self, *_):
        xdg_open(variables.git_wiki())

    def open_logs_directory(self, *_):
        run_task_once(OpenLogsDirectory, generic_already_running)

    def update_grapejuice(self, *_):
        if self._updating:
            return

        self._updating = True

        dialog("If the Grapejuice upgrade breaks your installation, please redo the Grapejuice installation according "
               "to the instructions in the Grapejuice git repository. The upgrade will begin after you close this "
               "dialog.")

        background.tasks.add(PerformUpdate(update_provider, reopen=True))

    def reinstall_grapejuice(self, *_):
        if self._updating:
            return

        self._updating = True
        background.tasks.add(PerformUpdate(update_provider, reopen=True))

    def uninstall_grapejuice(self, *_):
        do_it = yes_no_dialog("Uninstall Grapejuice", "Are you sure that you want to uninstall Grapejuice?")

        if not do_it:
            return

        params = uninstall.UninstallationParameters(
            yes_no_dialog(
                title="Remove Wineprefix?",
                message="Remove the Wineprefix that contains your installation of Roblox Studio? This will cause all "
                        "configuration of Roblox Studio to be permanently deleted! "
            ),
            for_reals=True
        )

        try:
            dialog("Grapejuice will now uninstall itself, the program will close when the process is finished.")
            uninstall.go(params)

        except Exception as e:
            msg = f"{e.__class__.__name__}: {str(e)}"
            LOG.error(msg)

            dialog(f"Failed to uninstall Grapejuice.\n\n{msg}")

    def show_background_task_errors(self, *_):
        errors = [*background_task_errors]
        background_task_errors.clear()

        dialog("Some errors occurred while running. The following dialogs will show them in order. Check the logs for "
               "the full details of the errors.")

        for i, err in enumerate(errors):
            dialog("Error " + str(i + 1) + ": " + str(err))

        on_background_errors_shown()


class MainWindow(WindowBase):
    def __init__(self):
        super().__init__(
            variables.grapejuice_glade(),
            MainWindowHandlers()
        )

        self._background_task_errors = []

        background.tasks.tasks_changed.add_listener(self.on_tasks_changed)
        on_destroy.add_listener(self.before_destroy)

        self.on_tasks_changed()
        self.background_task_errors_button.hide()
        self.set_update_status_visibility(False)

        if update_provider.can_update():
            self.reinstall_grapejuice_button.show()
            self.uninstall_button.show()

        else:
            self.reinstall_grapejuice_button.hide()
            self.uninstall_button.hide()

        on_background_task_error.add_listener(self.on_background_task_error)
        on_background_errors_shown.add_listener(self.on_background_task_errors_shown)

        self.perform_update_check()

    @property
    def window(self):
        return self.builder.get_object("main_window")

    @property
    def background_task_spinner(self):
        return self.builder.get_object("background_task_spinner")

    @property
    def background_task_errors_button(self):
        return self.builder.get_object("background_task_errors_button")

    @property
    def update_status(self):
        return self.builder.get_object("update_status")

    @property
    def update_status_label(self):
        return self.builder.get_object("update_status_label")

    @property
    def update_button(self):
        return self.builder.get_object("update_button")

    @property
    def uninstall_button(self):
        return self.builder.get_object("uninstall_button")

    @property
    def reinstall_grapejuice_button(self):
        return self.builder.get_object("reinstall_grapejuice_button")

    def set_update_status_visibility(self, visible: bool, ignore_elements: Iterable = None):
        for element in (self.update_status_label, self.update_button, self.update_status):
            if ignore_elements and element in ignore_elements:
                continue

            if visible:
                element.show()

            else:
                element.hide()

    def show_update_status(self, status: str, show_button: bool):
        if show_button:
            ignore_elements = []

        else:
            ignore_elements = [self.update_button]

        self.update_status_label.set_text(status)
        self.set_update_status_visibility(True, ignore_elements)

    def on_background_task_error(self, _task, e):
        self.background_task_errors_button.show()
        background_task_errors.append(e)

    def on_background_task_errors_shown(self):
        self.background_task_errors_button.hide()

    def on_tasks_changed(self):
        if background.tasks.count > 0:
            self.background_task_spinner.start()

        else:
            self.background_task_spinner.stop()

    def show(self):
        self.window.show()

    def before_destroy(self):
        background.tasks.tasks_changed.remove_listener(self.on_tasks_changed)

    def perform_update_check(self):
        if not update_provider.can_update():
            return

        w = self

        class CheckForUpdates(background.BackgroundTask):
            def __init__(self, **kwargs):
                super().__init__("Checking for a newer version of Grapejuice", **kwargs)

            def work(self) -> None:
                show_button = False

                if update_provider.update_available():
                    s = "This version of Grapejuice is out of date.\n" \
                        f"{update_provider.local_version()} -> {update_provider.target_version()}"
                    show_button = True

                else:
                    if update_provider.local_is_newer():
                        s = f"This version of Grapejuice is from the future\n{update_provider.local_version()}"

                    else:
                        s = f" Grapejuice is up to date\n{update_provider.local_version()}"

                if s:
                    w.update_status_label.set_text(s)
                    w.show_update_status(s, show_button)

        background.tasks.add(CheckForUpdates())

    def __del__(self):
        on_background_task_error.remove_listener(self.on_background_task_error)
        on_background_errors_shown.remove_listener(self.on_background_task_errors_shown)



#!/usr/bin/env python3

import getpass
import io
import json
import os
import re
import shutil
import subprocess
import sys
import time
import urllib.request
import uuid
from datetime import datetime
from typing import List, Union, Dict

TROUBLESHOOTER_VERSION = 9

TMP = os.path.join(os.path.sep, "tmp")
assert os.path.exists(TMP), "Fatal error: /tmp does not exist"
assert os.path.isdir(TMP), "Fatal error: /tmp is not a directory"

ORIGINAL_CWD = os.getcwd()
CWD = os.path.join(TMP, str(uuid.uuid4()))
WINEPREFIX_PATH = os.path.join(os.environ["HOME"], ".cache", "grapejuice-troubleshooter-wineprefix-{arch}")
USED_PREFIXES = set()

os.makedirs(CWD, exist_ok=True)
os.chdir(CWD)

CHECKLIST = list()
CHARSET = "UTF-8"
USERNAME = getpass.getuser()


def fatal_file():
    fp = open(os.path.join(TMP, "grapejuice-troubleshooter-fatal.log"), "a+")
    print(datetime.now(), file=fp)


class TextIOWrapperWrapper(io.TextIOWrapper):
    def write(self, __s: str) -> int:
        __s = __s.replace(USERNAME, "[REDACTED]")

        try:
            sys.stderr.write(__s)

        except Exception as e:
            with fatal_file() as fp:
                print(e.__class__.__name__, file=fp)
                print(e, file=fp)

        return super().write(__s)


OUTPUT_BUFFER = io.BytesIO()
TEXT_OUTPUT_BUFFER = TextIOWrapperWrapper(OUTPUT_BUFFER)


class Log:
    @staticmethod
    def info(*args, sep=" "):
        print("INFO: " + sep.join(list(map(str, args))), file=TEXT_OUTPUT_BUFFER)

    @staticmethod
    def error(*args, sep=" "):
        print("ERROR: " + sep.join(list(map(str, args))), file=TEXT_OUTPUT_BUFFER)


class CSVReport:
    _delimiter = ";"
    _header = ["Troubleshooting Function", "What are we checking?", "Status", "Fixes"]
    _rows = []

    @classmethod
    def add_row(cls, fun: str, what: str, status: bool, fixes: List[str]):
        if status:
            fixes = None

        cls._rows.append([fun, what, "PASS" if status else "FAIL", " :: ".join(fixes) if fixes else ""])

    @classmethod
    def empty_row(cls):
        cls.add_row("", "", True, [])

    @classmethod
    def to_string(cls):
        with io.TextIOWrapper(io.BytesIO()) as fp:
            print(cls._delimiter.join(cls._header), file=fp)
            for row in cls._rows:
                print(cls._delimiter.join(row), file=fp)

            fp.seek(0)
            s = fp.read()

        return s


VARS = dict()


def report_var(k, v):
    VARS[k] = str(v)
    Log.info(f"Reporting variable:", k, v)
    return v


def which(bin_name: str):
    if os.path.exists(bin_name):
        return os.path.abspath(bin_name)

    assert "PATH" in os.environ, "Your environment does not have $PATH. Your system is broken!"

    for path_dir in os.environ["PATH"].split(os.path.pathsep):
        path_dir: str = path_dir.strip()
        bin_path = os.path.join(path_dir, bin_name)

        if os.path.exists(bin_path) and not os.path.isdir(bin_path):
            return bin_path

    return None


class CommonFixes:
    wine32 = "Install a version of Wine that is capable of 32-bit support"
    fresh_wine = "Install a recent version of Wine (4.0 or higher)"
    wine64 = "Install a version of Wine that is capable of 64-bit support"
    c_tools = "Install C development tools for your distribution"
    follow_guide = "Follow the installation guide for your particular distribution." \
                   "https://gitlab.com/brinkervii/grapejuice/-/wikis/home"


def check(friendly_text: str, fixes: List[str] = None):
    def decorator(fn):
        def wrapper(*args, **kwargs) -> bool:
            Log.info("Performing check:", friendly_text)

            try:
                status = fn(*args, **kwargs)

                assert status is None or isinstance(status, bool), \
                    "Developer error: function does not return the right type\n" \
                    f"Expected bool or None, got {type(status)}"

                ok = status if isinstance(status, bool) else False

            except Exception as e:
                Log.error(e.__class__.__name__, e, sep=" ~ ")
                ok = False

                if os.environ.get("DEBUGGING", "0").strip() == "1":
                    raise e

            Log.info("OK: " if ok else "NOT_OK: ", friendly_text, fn.__name__, sep=" | ")

            CSVReport.add_row(fn.__name__, friendly_text, ok, fixes)

            return ok

        CHECKLIST.append(wrapper)

        return wrapper

    return decorator


class WinePrefix:
    _previous_env = dict()
    _wine: str
    _wine64: str
    _arch: str

    def __init__(self, arch: str = "win64", pfx: str = None, destroy_prefix: Union[bool, None] = False):
        self._wine = which("wine")
        self._wine64 = which("wine64")
        self._arch = arch

        if pfx is None:
            pfx = WINEPREFIX_PATH.format(arch=arch)

        if destroy_prefix is None:
            destroy_prefix = not os.path.exists(pfx)

        if destroy_prefix:
            USED_PREFIXES.add(pfx)

        self._set("WINEPREFIX", pfx)
        self._set("WINEARCH", self._arch)

    def _set(self, k, v):
        if k in os.environ:
            self._previous_env[k] = os.environ[k]

        else:
            self._previous_env[k] = False

        Log.info("Setting environment variable", k, v)

        os.environ[k] = v

    def __del__(self):
        for k, v in self._previous_env.items():
            if isinstance(v, str):
                Log.info("Restoring environment variable", k, v)
                os.environ[k] = v

            elif isinstance(v, bool):
                Log.info("Deleting environment variable", k)
                os.environ.pop(k)

            else:
                raise RuntimeError(f"Invalid environment variable type: {type(v)}")

    def run(self, cmd: List[str]) -> str:
        cmd = [self._wine, *cmd]

        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        try:
            while proc.returncode is None:
                proc.poll()
                time.sleep(0.1)

        except subprocess.CalledProcessError:
            pass

        output = proc.stdout.read().decode(CHARSET) if proc.stdout else "-- NO STDOUT --"
        err = proc.stderr.read().decode(CHARSET) if proc.stderr else "-- NO STDERR --"

        Log.info("Wine logs: ", err, "\n\n")

        return output


def c_compilers():
    compilers = ["gcc", "clang"]
    if "CC" in os.environ:
        compilers.append(os.environ["CC"])

    compilers = list(filter(lambda cc: not not cc, map(which, compilers)))
    return compilers


def pkg_config(pkg: str, cflags: bool = True, libs: bool = True):
    cmd = ["pkg-config", pkg]

    if cflags:
        cmd.append("--cflags")

    if libs:
        cmd.append("--libs")

    output = subprocess.check_output(cmd)
    output = output.decode(CHARSET).strip()

    Log.info(cmd, output, sep=" -> ")

    return output


@check("Log troubleshooter version")
def log_troubleshooter_version():
    report_var("Troubleshooter Version", TROUBLESHOOTER_VERSION)

    return True


@check("Log directory info")
def log_cwd():
    report_var("ORIGINAL_CWD", ORIGINAL_CWD)
    report_var("CWD", CWD)

    assert os.path.isdir(ORIGINAL_CWD), "The original CWD is not a directory!"
    assert os.path.isdir(CWD), "The current working directory is not a directory?!!?!!one"

    return True


@check("Are we running in Python 3?", fixes=["Run the script with Python3"])
def is_python3():
    report_var("Version info", sys.version_info)
    report_var("Version", sys.version)
    report_var("API Version", sys.api_version)

    return sys.version_info.major == 3


@check("Are we running in at least python 3.7?", fixes=["Install a version of Python that is at least 3.7"])
def have_python37():
    return sys.version_info.major == 3 and sys.version_info.minor >= 7


UNAME = which("uname")
if UNAME is not None:
    Log.info("Found uname:", UNAME)


    @check("Are we on linux?", fixes=["Use the OS Grapejuice was meant for"])
    def on_linux():
        out = subprocess.check_output([UNAME]).decode(CHARSET).strip()
        report_var("Kernel", out)
        answer = out.lower().startswith("linux")

        if answer:
            out = subprocess.check_output([UNAME, "-r"]).decode(CHARSET).strip()
            report_var("Kernel version", out)

        return answer

else:
    Log.error("Could not find uname")

OS_RELEASE = os.path.join(os.path.sep, "etc", "os-release")


@check("Do we have the OS release file on the system?", fixes=["Run Grapejuice on an LSB compatible distribution."])
def have_os_release():
    exists = os.path.exists(OS_RELEASE)
    if exists:
        with open(OS_RELEASE, "r") as fp:
            contents = fp.read()
            Log.info("OS Release contents:\n", contents)

            for line in contents.split("\n"):
                if not line:
                    continue

                s = line.split("=")
                k = s[0]
                v = "=".join(s[1:])

                report_var("OS_RELEASE_" + k, v)

    return exists


LSCPU = which("lscpu")

if LSCPU is not None:
    Log.info("Found lscpu:", LSCPU)


    @check("Do we have a compatible CPU?")
    def have_compatible_cpu():
        ptn = re.compile(r"Architecture:\s*(.*)")
        out = subprocess.check_output([LSCPU]).decode(CHARSET).strip()

        Log.info("CPU Info:\n", out)

        for line in out.split("\n"):
            match = ptn.search(line)
            if match:
                arch = match.group(1).strip()
                report_var("CPU Architecture", arch)
                return arch == "x86_64"

        return False

else:
    Log.info("lscpu could not be found, so we cannot check for the CPU architecture")


@check("Do we have a compiler?", fixes=[CommonFixes.follow_guide, "Install a compatible C compiler"])
def have_compiler():
    compilers = c_compilers()
    report_var("C Compilers", os.path.pathsep.join(compilers))
    return len(compilers) > 0


@check("Do we have pip or pip3?", fixes=[CommonFixes.follow_guide, "Install pip for Python 3"])
def have_pip():
    pip = which("pip")
    pip3 = which("pip3")

    report_var("pip", pip)
    report_var("pip3", pip3)

    return (pip or pip3) is not None


@check("Do we have pkg-config?", fixes=[CommonFixes.follow_guide, CommonFixes.c_tools])
def have_pkg_config():
    return report_var("PKG_CONFIG", which("pkg-config")) is not None


@check("Are the native libraries for cairo installed?",
       fixes=[CommonFixes.follow_guide, "Install the cairo development packages for your distribution"])
def have_cairo_natives():
    return report_var("PKG_CONFIG_CAIRO", pkg_config("cairo")) is not None


@check("Can we update the GTK icon cache?",
       fixes=[CommonFixes.follow_guide, "Install the GTK+ icon utilities for your distribution"])
def can_update_icon_cache():
    return report_var("GTK icon cache command", which("gtk-update-icon-cache")) is not None


@check("Can we update the desktop file database?",
       fixes=[CommonFixes.follow_guide, "Install the freedesktop database utilities for your distribution"])
def can_update_desktop_database():
    return report_var("Update desktop database command", which("update-desktop-database")) is not None


@check("Do we have XDG mime?", fixes=[CommonFixes.follow_guide, "Install the XDG utilities for your distribution"])
def have_xdg_mime():
    return report_var("XDG Mime command", which("xdg-mime")) is not None


@check("Do have have GObject?",
       fixes=[CommonFixes.follow_guide, "Install the GObject development packages for your distribution"])
def have_gobject():
    return report_var("PKG_CONFIG_GOBJECT", pkg_config("gobject-2.0")) is not None


@check("Do we have GObject introspection?",
       fixes=[CommonFixes.follow_guide, "Install the GObject introspection development packages for your distribution"])
def have_gobject_introspection():
    return report_var("PKG_CONFIG_GOBJECT_INTROSPECTION", pkg_config("gobject-introspection-1.0")) is not None


@check("Do we have git?", fixes=[CommonFixes.follow_guide, "Install git"])
def have_git():
    return report_var("git binary path", which("git")) is not None


@check("Do we have wine?", fixes=[CommonFixes.follow_guide, "Install Wine"])
def have_wine():
    return report_var("Wine binary path", which("wine")) is not None


@check("Do we have 64-bit wine?", fixes=[CommonFixes.follow_guide, CommonFixes.wine64])
def have_wine64():
    return report_var("Wine64 binary path", which("wine64")) is not None


@check("Log Wine version")
def print_wine_version():
    version = subprocess.check_output(["wine", "--version"]).decode(CHARSET).strip()
    report_var("WINE_VERSION", version)
    Log.info(version)

    return not not version


HOME = os.environ.get("HOME", None)


@check("Do we have the $HOME environment variable?")
def have_home_variable():
    return HOME is not None


if have_home_variable():
    GRAPEJUICE_WINEPREFIX = os.path.join(HOME, ".local", "share", "grapejuice", "wineprefix")


    @check("Do we have a wineprefix?")
    def have_wineprefix():
        return os.path.exists(GRAPEJUICE_WINEPREFIX)


    @check("Do we have a valid wineprefix?")
    def have_valid_wineprefix():
        if have_wineprefix():
            return os.path.isdir(GRAPEJUICE_WINEPREFIX)

        return False


    if have_wineprefix() and have_valid_wineprefix():
        class RegistryFile:
            _path: str
            _sections: Dict[str, Dict]
            _current_section: Dict[str, str] = None
            _section_header_ptn = re.compile(r"\[(.*)?].*")
            _key_ptn = re.compile(r"\"([\w\d]+)?\"\s*=\s*(.*)")

            def __init__(self, path: str):
                self._path = path
                self._sections = dict()

                if self.exists:
                    self._parse()

            @property
            def exists(self):
                return os.path.exists(self._path)

            @property
            def sections(self) -> List[str]:
                return list(self._sections.keys())

            def get_keys(self, section_key: str) -> Dict[str, str]:
                return self._sections[section_key]

            def _parse(self):
                with open(self._path, "rb") as fp:
                    content = fp.read()
                    text = content[2:].decode("iso8859-1").replace("\0", "")

                for line in text.split("\n"):
                    line = line.strip()

                    if not line:
                        continue

                    if line.startswith("Windows"):
                        continue

                    match = self._section_header_ptn.match(line)
                    if match:
                        self._current_section = self._sections.setdefault(match.group(1), {})
                        continue

                    match = self._key_ptn.match(line)
                    if match and (self._current_section is not None):
                        self._current_section[match.group(1).strip()] = match.group(2).strip()

            def __str__(self):
                return self._path


def run_wine_test_command(prefix: WinePrefix):
    out = prefix.run(["cmd", "/c", "echo Hello from Wine"])
    return prefix, out


@check("Can we make a valid 32-bit wineprefix?", fixes=[CommonFixes.wine32])
def can_make_valid_32_bit_prefix():
    prefix, out = run_wine_test_command(WinePrefix(arch="win32", destroy_prefix=True))
    Log.info(out)
    del prefix

    return True


@check("Can we make a valid 64-bit wineprefix?", fixes=[CommonFixes.wine64])
def can_make_valid_64_bit_prefix():
    prefix, out = run_wine_test_command(WinePrefix(arch="win64", destroy_prefix=True))
    Log.info(out)
    del prefix

    error_ptn = re.compile(r".*?:\s*WINEARCH set to win64 but .*?is a 32-bit installation.*")
    assert error_ptn, "Couldn't compile the pattern that checks for a 64-bit error"

    match = error_ptn.search(out)
    if match is None:
        return True

    return False


@check("Can we access the Grapejuice package?", fixes=[CommonFixes.follow_guide])
def have_grapejuice_package():
    try:
        from grapejuice import __version__
        Log.info("Grapejuice version:", __version__)

        return True

    except ImportError as e:
        Log.error(e)
        return False


HASTEBIN_DOCUMENTS = "https://hastebin.com/documents"
HASTEBIN_FILE = "https://hastebin.com/{file_id}"


def strict_yes_no(msg: str):
    querying = True
    result = False

    while querying:
        response = input(f"{msg} Type yes in capital letters if you do, if not type 'n' or 'no'.\n> ").strip()

        if response == "YES":
            querying = False
            result = True

        elif response.lower() == "n" or response.lower() == "no":
            querying = False
            result = False

        else:
            print("That is an invalid answer, please try again")
            print("")

    return result


def yes_no(msg: str):
    querying = True
    result = False

    def print_invalid():
        print("That is an invalid answer, please try again")
        print("")

    while querying:
        response = input(f"{msg} [y/n]\n> ").strip().lower()

        if response == "y" or response == "n":
            querying = False

            if response == "y":
                result = True

            elif response == "n":
                result = False

            else:
                print_invalid()

        else:
            print_invalid()

    return result


def display_file(path):
    cmds = ["less", "more", "cat"]

    if "EDITOR" in os.environ:
        cmds.insert(0, os.environ["EDITOR"])

    if "PAGER" in os.environ:
        cmds.insert(0, os.environ["PAGER"])

    for cmd in map(which, cmds):
        if cmd:
            subprocess.call([cmd, path])
            print("")
            return


def str_to_hastebin(s: str):
    req = urllib.request.Request(
        HASTEBIN_DOCUMENTS,
        method="POST",
        data=s.encode(CHARSET),
        headers={
            "User-Agent": f"Grapejuice Troubleshooter (CLI; Linux; {TROUBLESHOOTER_VERSION})",
            "Content-Type": "application/json",
            "DNT": 1,
            "Accept": "application/json, text/plain"
        }
    )

    response = urllib.request.urlopen(req)
    response_text: str = response.read().decode(CHARSET)
    response_object = json.loads(response_text)

    return HASTEBIN_FILE.format(file_id=response_object["key"])


def main():
    n_passed = sum(map(lambda f: f(), CHECKLIST))
    Log.info(f"{n_passed}/{len(CHECKLIST)} passed")

    TEXT_OUTPUT_BUFFER.seek(0)
    report = TEXT_OUTPUT_BUFFER.read()

    CSVReport.empty_row()
    CSVReport.empty_row()
    CSVReport.add_row("KEY", "VALUE", True, [])

    for k, v in VARS.items():
        CSVReport.add_row(k, v.replace(USERNAME, "[REDACTED]"), True, [])

    report_path = "/tmp/grapejuice-report.csv"

    report_fp = io.TextIOWrapper(io.BytesIO())

    print(CSVReport.to_string(), file=report_fp)
    print("", file=report_fp)
    print("""
##################
### LOG OUTPUT ###
##################
    """, file=report_fp)
    print(report, file=report_fp)

    report_fp.seek(0)
    with open(report_path, "w+") as fp:
        fp.write(report_fp.read())

    for pfx in USED_PREFIXES:
        if os.path.exists(pfx) and os.path.isdir(pfx):
            Log.info("Removing wineprefix", pfx)
            shutil.rmtree(pfx)

    if os.path.exists(CWD):
        Log.info("Removing", CWD)

        os.chdir(ORIGINAL_CWD)
        shutil.rmtree(CWD)

    for _ in range(30):
        print("")

    return_code = 0 if n_passed == len(CHECKLIST) else 1
    if return_code:
        print("Some checks did not pass, please review the report to see what's not working correctly")
        print("")

    else:
        print("All checks passed!")
        print("")

    print(f"Saved report to: {report_path}")
    print("If you intend on sharing the report, please make sure that it generated correctly.")
    print("")

    if yes_no("Do you want to review the report?"):
        display_file(report_path)

    if strict_yes_no("Do you want to upload your report to Hastebin, so you can share it with people that are helping "
                     "you troubleshoot?"):
        report_fp.seek(0)
        url = str_to_hastebin(report_fp.read())
        print("")
        print("Hastebin url for you to share: " + url)

    if return_code == 0:
        print("")
        print("Everything was A-OK, have a nice day!")

    sys.exit(return_code)


main()



import json
import logging
import os
from typing import List, Iterable, Dict

import requests

LOG = logging.getLogger(__name__)


class RobloxApplications:
    studio_app = "StudioApp"
    pc_desktop_client = "PCDesktopClient"


def download_fast_flags(app: str):
    response = requests.get(
        "https://clientsettingscdn.roblox.com/v1/settings/application",
        params={
            "applicationName": app
        }
    )

    try:
        response.raise_for_status()
        return response.json()["applicationSettings"]

    except requests.HTTPError as e:
        LOG.error(f"Could not get fast flags for applicaton {app}: {e}")
        return {}


def mangle_flags(flags: Dict[str, str]):
    new_flags = dict()

    for k, v in flags.items():
        if isinstance(v, str):
            v = v.split(";")[0].strip()
            v_lower = v.lower()

            if v_lower in ("true", "false"):
                v = v_lower == "true"

            elif v_lower.isnumeric():
                v = int(v)

        new_flags[k] = v

    return new_flags


class FastFlag:
    def __init__(self, name, value):
        self._name = name
        self._original_value = value
        self.value = value

    def is_a(self, cls):
        return isinstance(self.value, cls)

    @property
    def name(self):
        return self._name

    @property
    def has_changed(self):
        return self.value != self._original_value

    def to_tuple(self):
        return self.name, self.value

    def reset(self):
        self.value = self._original_value

    def __lt__(self, other):
        if isinstance(other, FastFlag):
            return self._name < other.name

        return -1

    def __repr__(self):
        return "FFlag '{}': {}".format(self._name, self.value)


class FastFlagList:
    def __init__(self, initial_values: Iterable[FastFlag] = None):
        if initial_values is None:
            self._list: List[FastFlag] = list()

        else:
            self._list = list(initial_values)

    def clear(self):
        self._list = list()

    def import_file(self, fast_flags_path):
        with open(fast_flags_path, "r") as fp:
            json_object = json.load(fp)
            return self.import_dict(json_object)

    def import_dict(self, fast_flags):
        self._list = list(map(lambda t: FastFlag(*t), fast_flags.items()))

        self.sort()

        return self

    def export_to_file(self, fast_flags_path):
        os.makedirs(os.path.dirname(fast_flags_path), exist_ok=True)

        with open(fast_flags_path, "w+") as fp:
            json.dump(self.to_dict(), fp)

    def overlay_flags(self, other_flags):
        d = dict(zip(map(lambda f: f.name, self), self._list))

        for flag in filter(lambda f: f.name in d, other_flags):
            d[flag.name].value = flag.value

        self.sort()

    def get_changed_flags(self):
        return FastFlagList(initial_values=filter(lambda flag: flag.has_changed, self._list))

    def to_dict(self):
        return dict(map(lambda flag: flag.to_tuple(), self))

    def reset_all_flags(self):
        for flag in self:
            flag.reset()

    def sort(self):
        changed_flags = list(filter(lambda f: f.has_changed, self._list))
        unchanged_flags = list(filter(lambda f: not f.has_changed, self._list))

        changed_flags.sort()
        unchanged_flags.sort()

        self._list = changed_flags + unchanged_flags

    def __iter__(self):
        for flag in self._list:
            yield flag

    def __len__(self):
        return len(self._list)

    def __getitem__(self, *args):
        return self._list.__getitem__(*args)



import atexit
import logging
import os
import re
import signal

import psutil

LOG = logging.getLogger(__name__)


class EmptyPIDError(RuntimeError):
    def __init__(self, path):
        super().__init__("Got empty string from PID file " + path)


class AlreadyRunningError(RuntimeError):
    pass


class PIDFile:
    def __init__(self, name: str):
        self._name = re.sub(r"(\W)", "_", name)

        if "XDG_RUNTIME_DIR" in os.environ:
            xdg_runtime_dir = os.environ["XDG_RUNTIME_DIR"]
            LOG.debug(f"PIDFile instance {self._name} is using XDG_RUNTIME_DIR '{xdg_runtime_dir}")
            self._path = os.path.join(xdg_runtime_dir, self._name + ".pid")

        else:
            LOG.debug(f"PIDFile instance {self._name} is using /tmp")
            self._path = os.path.join("/tmp", self._name + ".pid")

        self._wrote_pid = False
        LOG.info(f"PIDFile {self._name} got placed at '{self._path}'")

        atexit.register(self._at_exit)

    def _at_exit(self, *_):
        self._remove_file()

    def exists(self):
        return os.path.exists(self._path)

    @property
    def pid(self):
        with open(self._path, "r") as fp:
            s = fp.read().strip()
            if not s:
                raise EmptyPIDError(self._path)

            return int(s)

    def _remove_file(self):
        if self._wrote_pid:
            if os.path.exists(self._path):
                LOG.debug(f"Removing pidfile {self._path}")
                os.remove(self._path)

            else:
                LOG.warning(f"PIDFile {self._path} does not exist!")

        else:
            LOG.debug("Did not write PID, therefore I am not removing it")

    def is_running(self, remove_junk=True):
        if not self.exists():
            return False

        try:
            process = psutil.Process(pid=self.pid)
            return process.is_running()

        except psutil.NoSuchProcess:
            if remove_junk:
                self._remove_file()

            return False

    def write_pid(self):
        try:
            if self.is_running():
                raise AlreadyRunningError

        except EmptyPIDError:
            pass

        with open(self._path, "w+") as fp:
            fp.write(str(os.getpid()))

        self._wrote_pid = True

    def kill(self):
        if self.is_running():
            os.kill(self.pid, signal.SIGINT)
            self._remove_file()


def daemon_pid_file():
    return PIDFile("grapejuiced")



import logging
import threading
import traceback
from typing import Union, Callable

from grapejuice_common.util.event import Event

OptionalCallback = Union[Callable, None]


class Task:
    _log: logging.Logger
    _on_finish_callback: OptionalCallback = None
    _on_error_callback: OptionalCallback = None

    def __init__(
        self, name,
        on_finish_callback: OptionalCallback = None,
        on_error_callback: OptionalCallback = None
    ):
        self._log = logging.getLogger(f"{Task.__name__}/{name}")

        self._on_finish_callback = on_finish_callback
        self._on_error_callback = on_error_callback
        self._name = name
        self._finished = False
        self._collection: Union[None, TaskCollection] = None

    @property
    def finished(self):
        return self._finished

    @property
    def collection(self):
        return self._collection

    @collection.setter
    def collection(self, value):
        assert self._collection is None, "Can only set collection once"
        self._collection = value

    def on_finished(self):
        if callable(self._on_finish_callback):
            self._on_finish_callback(self)

    def on_error(self, e: Exception):
        self._log.error(str(e))

        try:
            self._log.error(traceback.format_exc())

        except Exception as format_error:
            # There was an error while you were formatting your error
            # So let's error while erroring
            # ...dawg
            self._log.error(str(format_error))

        if callable(self._on_error_callback):
            self._on_error_callback(self, e)

    def finish(self):
        self.on_finished()
        self._finished = True

        if self.collection is not None:
            self.collection.remove(self)

    def work(self):
        pass

    @property
    def name(self):
        return self._name


class BackgroundTask(threading.Thread, Task):
    def __init__(self, name="Untitled task", **kwargs):
        threading.Thread.__init__(self)
        Task.__init__(self, name, **kwargs)

    def run(self) -> None:
        try:
            self.work()

        except Exception as e:
            self.on_error(e)

        self.finish()

    def __repr__(self):
        return "BackgroundTask: {}".format(self._name)


class MockBackgroundTask(Task):
    def start(self):
        pass


class TaskCollection:
    def __init__(self):
        self._tasks = []

        self.task_added = Event()
        self.task_removed = Event()
        self.tasks_changed = Event()

    def add(self, task: BackgroundTask):
        task.collection = self
        self._tasks.append(task)
        task.start()

        self.task_added(task)
        self.tasks_changed()

    def remove(self, task):
        if task in self._tasks:
            self._tasks.remove(task)

        self.task_removed(task)
        self.tasks_changed()

    @property
    def count(self):
        return len(self._tasks)


tasks = TaskCollection()



import atexit
import logging
import os
import re
import shutil
import signal
import subprocess
import sys
import time
from datetime import datetime
from pathlib import Path
from string import Template
from subprocess import DEVNULL
from typing import List, Union

from grapejuice_common import variables
from grapejuice_common.logs.log_util import log_on_call, log_function

LOG = logging.getLogger(__name__)

space_version_ptn = re.compile(r"wine-(.+?)\s+")
non_space_version_ptn = re.compile(r"wine-(.+)")
space = " "

open_fds = []


class ProcessWrapper:
    on_exit: callable = None

    def __init__(self, proc: subprocess.Popen, on_exit: callable = None):
        self.proc = proc
        self.on_exit = on_exit

    @property
    def exited(self):
        proc = self.proc

        if proc.returncode is None:
            proc.poll()

        return proc.returncode is not None

    def kill(self):
        if not self.exited:
            os.kill(self.proc.pid, signal.SIGINT)

    def __del__(self):
        del self.proc


processes: List[ProcessWrapper] = []
is_polling = False


@log_on_call("Preparing for Wine")
def prepare():
    from grapejuice_common.features.settings import current_settings
    from grapejuice_common.features import settings

    prefix_dir = variables.wineprefix_dir()

    user_env = current_settings.get(settings.k_environment_variables, {})
    apply_env = {
        "WINEDLLOVERRIDES": current_settings.get(settings.k_dll_overrides),
        **user_env,
        "WINEPREFIX": prefix_dir,
        "WINEARCH": "win64"
    }

    # Variables in os.environ take priority
    for k, v in user_env.items():
        apply_env[k] = os.environ.get(k, v)

    # Wine generates giant logs for some people
    # Setting WINEDEBUG to -all *should* fix it
    if "WINEDEBUG" not in apply_env:
        apply_env["WINEDEBUG"] = "-all"

    # Apply env
    for k, v in apply_env.items():
        os.environ[k] = v

    if not os.path.exists(prefix_dir):
        os.makedirs(prefix_dir)


@log_on_call("Running Wine configuration")
def winecfg():
    prepare()
    run_exe("winecfg", run_async=True)


@log_on_call("Running registry editor")
def regedit():
    prepare()
    run_exe("regedit", run_async=True)


@log_on_call("Running Windows Explorer")
def explorer():
    prepare()
    run_exe("explorer", run_async=True)


def load_registry_file(srcfile, prepare_wine: bool = True):
    LOG.info(f"Loading registry file {srcfile} into the wineprefix")

    if prepare_wine:
        prepare()

    target_filename = str(int(time.time())) + ".reg"
    target_path = os.path.join(variables.wine_temp(), target_filename)
    shutil.copyfile(srcfile, target_path)

    winreg = "C:\\windows\\temp\\{}".format(target_filename)
    run_exe("regedit", "/S", winreg, run_async=False, use_wine64=False)
    run_exe("regedit", "/S", winreg, run_async=False, use_wine64=True)

    os.remove(target_path)


def load_patched_registry_files(source: Path, patches: dict = None):
    prepare()

    target_filename = str(int(time.time())) + ".reg"
    target_path = Path(variables.wine_temp()) / target_filename

    with source.open("r") as fp:
        template = Template(fp.read())

    with target_path.open("w+") as fp:
        fp.write(template.safe_substitute(patches))

    winreg = "C:\\windows\\temp\\{}".format(str(target_filename))
    run_exe("regedit", "/S", winreg, run_async=False, use_wine64=False)
    run_exe("regedit", "/S", winreg, run_async=False, use_wine64=True)

    os.remove(target_path)


@log_on_call("Running Winetricks")
def wine_tricks():
    prepare()

    processes.append(ProcessWrapper(
        subprocess.Popen(["winetricks"])
    ))

    poll_processes()


@log_on_call("Disabling MIME associations in wineprefix")
def disable_mime_assoc():
    load_registry_file(os.path.join(variables.assets_dir(), "disable_mime_assoc.reg"))


def set_roblox_document_path():
    src_path = Path(variables.assets_dir()) / "roblox_documents_folder.reg"
    patches = dict()

    documents_dir = "Z:" + variables.xdg_documents().replace("/", "\\\\")
    patches["DOCUMENTS_DIR"] = documents_dir
    LOG.info(f"Setting the roblox documents directory to '{documents_dir}'")

    load_patched_registry_files(src_path, patches)


@log_on_call("Sandboxing user directories in the wineprefix")
def sandbox():
    user_dir = variables.wine_user()

    if os.path.exists(user_dir) and os.path.isdir(user_dir):
        for file in os.listdir(user_dir):
            p = os.path.join(user_dir, file)

            if os.path.islink(p):
                LOG.info(f"Sandboxing {file}")
                os.remove(p)
                os.makedirs(p, exist_ok=True)


@log_on_call("Configuring the wineprefix")
def configure_prefix():
    disable_mime_assoc()
    sandbox()
    set_roblox_document_path()


@log_on_call("Creating the wineprefix")
def create_prefix():
    configure_prefix()


@log_function
def prefix_exists():
    return os.path.exists(variables.wineprefix_dir())


@log_function
def run_exe_no_daemon(
    command: List[str],
    exe_name: str,
    run_async: bool,
    post_run_function: callable = None
) -> Union[ProcessWrapper, None]:
    LOG.info("Running in no_daemon_mode")

    log_dir = Path(variables.logging_directory())
    os.makedirs(log_dir, exist_ok=True)

    LOG.info("Opening log fds")

    ts = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    stdout_path = log_dir / f"{ts}_{exe_name}_stdout.log"
    stderr_path = log_dir / f"{ts}_{exe_name}_stderr.log"

    stdout_fd = stdout_path.open("wb+")
    stderr_fd = stderr_path.open("wb+")

    open_fds.extend((stdout_fd, stderr_fd))

    if run_async:
        LOG.info("Running process asynchronously")

        wrapper = ProcessWrapper(
            subprocess.Popen(
                command,
                stdout=stdout_fd,
                stderr=stderr_fd
            ),
            on_exit=post_run_function
        )

        processes.append(wrapper)
        poll_processes()

        return wrapper

    else:
        LOG.info("Running process synchronously")

        subprocess.call(
            command,
            stdout=stdout_fd,
            stderr=stderr_fd
        )

        if callable(post_run_function):
            post_run_function()

        return None


@log_function
def run_exe_in_daemon(command: List[str], post_run_function: callable = None) -> ProcessWrapper:
    LOG.info("Running process for daemon mode")

    p = subprocess.Popen(command, stdin=DEVNULL, stdout=sys.stdout, stderr=sys.stderr)
    wrapper = ProcessWrapper(p, on_exit=post_run_function)

    processes.append(wrapper)
    poll_processes()

    return wrapper


@log_function
def run_exe(
    exe_path: Union[Path, str],
    *args,
    run_async=False,
    use_wine64=False,
    post_run_function: callable = None
) -> Union[ProcessWrapper, None]:
    from grapejuice_common.features.settings import current_settings
    from grapejuice_common.features import settings

    prepare()
    LOG.info("Prepared Wine.")

    if isinstance(exe_path, Path):
        exe_path_string = str(exe_path.resolve())
        exe_name = exe_path.name

    elif isinstance(exe_path, str):
        exe_path_string = exe_path
        exe_name = exe_path.split(os.path.sep)[-1]

    else:
        raise ValueError(f"Invalid value type for exe_path: {type(exe_path)}")

    LOG.info(f"Resolved exe path to {exe_path_string}")

    wine_binary = variables.wine_binary_64() if use_wine64 else variables.wine_binary()
    command = [wine_binary, exe_path_string, *args]

    if current_settings.get(settings.k_no_daemon_mode):
        return run_exe_no_daemon(command, exe_name, run_async, post_run_function=post_run_function)

    else:
        return run_exe_in_daemon(command, post_run_function=post_run_function)


def _poll_processes() -> bool:
    """
    Makes sure zombie launchers are taken care of
    :return: Whether or not processes remain
    """
    global is_polling
    exited = []

    for proc in processes:
        if proc.exited:
            exited.append(proc)

            if proc.proc.returncode != 0:
                LOG.error(f"Process returned with non-zero exit code {proc.proc.returncode}")

    for proc in exited:
        processes.remove(proc)

        if callable(proc.on_exit):
            proc.on_exit()

        del proc

    processes_left = len(processes) > 0
    if not processes_left:
        is_polling = False
        LOG.info("No processes left to poll, exiting thread")

    return processes_left


def poll_processes():
    global is_polling

    if is_polling:
        return

    LOG.info("Starting polling thread")

    from gi.repository import GObject
    GObject.timeout_add(100, _poll_processes)


def close_fds(*_, **__):
    LOG.info("Closing fds")

    for fd in open_fds:
        fd.close()

    open_fds.clear()


atexit.register(close_fds)



import logging
from typing import List, Tuple


class TaskError(RuntimeError):
    def __init__(self, message: str, can_continue: bool = True):
        super().__init__(message)
        self._can_continue = can_continue

    @property
    def can_continue(self) -> bool:
        return self._can_continue


class TaskSequence:
    _ok_str = "Y"
    _fail_str = "N"
    _pending_str = "?"

    def __init__(self, name: str):
        self._name = name
        self._log = logging.getLogger(self._name)
        self._tasks: List[Tuple[str, callable]] = []
        self._results: List[str] = []
        self._reset_results()

    def task(self, name: str):
        tasks = self._tasks

        def wrap_function(func: callable):
            tasks.append((name, func))

            def wrapper():
                return func()

            return wrapper

        return wrap_function

    def _reset_results(self):
        self._results = ["?" for _ in self._tasks]

    def _log_results(self):
        self._log.info("Task status: " + "".join(self._results))

    def run(self):
        self._log.info("Running task sequence")
        self._reset_results()
        self._log_results()

        task_counter = 0
        for task_name, task_func in self._tasks:
            self._log.info(f"-- Running task -> {task_name} --")

            try:
                task_func(self._log)
                self._results[task_counter] = self._ok_str

            except TaskError as e:
                self._log.error(str(e))
                self._results[task_counter] = self._fail_str

                if e.can_continue:
                    self._log.warning(f"Task '{task_name}' failed, but we can continue")

                else:
                    self._log.error(f"Aborting due to failed task '{task_name}'")
                    raise e

            self._log.info("-- End of task --")

            task_counter += 1
            self._log_results()

        self._log.info(f"{self._name} OK")



import subprocess
import sys

from grapejuice_packaging.builders.package_builder import PackageBuilder


class PyPiPackageBuilder(PackageBuilder):
    def build(self):
        self.clean_build()

    def dist(self):
        self.clean_dist()
        subprocess.check_call([sys.executable, "setup.py", "sdist", "bdist_wheel"])



import getpass
import logging
import os
import sys
from datetime import datetime
from logging import LogRecord
from pathlib import Path
from typing import IO, Union, Optional

from grapejuice_common import variables


def _strip_pii(s: str):
    s = s.replace(variables.home(), "~")

    if getpass.getuser().lower() != "root":
        s = s.replace(getpass.getuser(), "[REDACTED]")

    return s


class GrapejuiceLogFormatter(logging.Formatter):

    def format(self, record: LogRecord) -> str:
        return _strip_pii(super().format(record))

    def formatTime(self, record: LogRecord, datefmt: Optional[str] = ...) -> str:
        return _strip_pii(super().formatTime(record, datefmt))

    def formatException(self, ei) -> str:
        return _strip_pii(super().formatException(ei))

    def formatMessage(self, record: LogRecord) -> str:
        return _strip_pii(super().formatMessage(record))

    def formatStack(self, stack_info: str) -> str:
        return _strip_pii(super().formatStack(stack_info))


class LoggerConfiguration:
    _output_stream: IO = sys.stdout
    _output_file: Union[str, Path] = None
    _formatter: logging.Formatter = None
    _environment_key: Union[str, None] = "LOG_LEVEL"
    _log_level_override: Union[str, None] = None

    def __init__(self, app_name: str):
        self._app_name = app_name
        self._formatter = GrapejuiceLogFormatter(f"[%(levelname)s] {app_name}/%(name)s:- %(message)s")

    @property
    def use_output_stream(self):
        return self._output_stream is not None

    @property
    def output_stream(self):
        stream = self._output_stream
        assert stream is not None
        return stream

    @output_stream.setter
    def output_stream(self, stream: IO):
        self._output_stream = stream

    @property
    def use_output_file(self):
        return self._output_file is not None

    @property
    def output_file(self) -> str:
        return self._output_file

    @output_file.setter
    def output_file(self, path: str):
        self._output_file = path

        parent_file = os.path.dirname(path)
        os.makedirs(parent_file, exist_ok=True)

    @property
    def formatter(self) -> logging.Formatter:
        return self._formatter

    @property
    def log_level_str(self) -> str:
        if self._log_level_override is not None:
            return self._log_level_override.upper()

        if self._environment_key is not None and self._environment_key in os.environ:
            return os.environ[self._environment_key].upper()

        return "INFO"

    @property
    def app_name(self):
        return self._app_name


def configure_logging(app_name: str = None, configuration: LoggerConfiguration = None):
    if configuration is None:
        assert isinstance(app_name, str)
        configuration = LoggerConfiguration(app_name)

        datetime_now = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        configuration.output_file = \
            os.path.join(variables.logging_directory(), f"{datetime_now}_{configuration.app_name}.log")

    root_logger = logging.getLogger()

    if configuration.use_output_stream:
        stream_handler = logging.StreamHandler(configuration.output_stream)
        stream_handler.setFormatter(configuration.formatter)
        root_logger.addHandler(stream_handler)

    if configuration.use_output_file:
        file_handler = logging.FileHandler(configuration.output_file, "w+", "UTF-8")
        file_handler.setFormatter(configuration.formatter)
        root_logger.addHandler(file_handler)

    log_level = configuration.log_level_str
    assert hasattr(logging, log_level), \
        f"An invalid log level string was provided: {log_level}"

    root_logger.setLevel(getattr(logging, log_level))

    root_logger.info(f"Log level was set to '{log_level}'")
    if configuration.use_output_file:
        root_logger.info(f"The log file is stored at '{configuration.output_file}'")



from grapejuice import __version__ as gj_package_version

project_name = "Grapejuice"
package_name = "grapejuice"
package_version = gj_package_version
package_license = "GPL-3"
package_platform = "Linux"
package_description = "A simple wine+roblox management application"
package_repository = "https://gitlab.com/brinkervii/grapejuice"
author_name = "BrinkerVII"
author_email = "brinkervii@gmail.com"



import json
import logging
import os
import shutil
import subprocess
import sys
import tarfile
from io import BytesIO
from pathlib import Path
from string import Template

import grapejuice.__about__ as about
import grapejuice_common.variables as v
import grapejuice_packaging.packaging_resources as res
from grapejuice_common.util.task_sequence import TaskSequence
from grapejuice_packaging.builders.package_builder import PackageBuilder

LOG = logging.getLogger(__name__)


class LinuxPackageConfiguration:
    root: str = None
    level_1_directory: str = "usr"
    copy_packages: bool = True
    python_site_type: str = "dist-packages"
    python_site_version: str = "python3"
    package_name: str = f"{about.package_name}-{about.package_version}.tar.gz"
    target_system_root: str = os.path.sep

    def __init__(self, root: str):
        self.root = root


def _build_package(configuration: LinuxPackageConfiguration):
    root = configuration.root
    build = TaskSequence("Build Linux Package")

    bin_path_components = [configuration.level_1_directory, "bin"]
    grapejuice_exec_name = "grapejuice"

    if configuration.copy_packages:
        python_site = Path(
            root,
            configuration.level_1_directory,
            "lib",
            configuration.python_site_version,
            configuration.python_site_type
        )

        @build.task("Copy packages to site")
        def copy_packages(log):
            log.info(f"Using site directory: {python_site}")
            os.makedirs(python_site, exist_ok=True)

            subprocess.check_call([
                sys.executable, "-m", "pip",
                "install", ".",
                "--no-dependencies",
                "--target", str(python_site)
            ])

    @build.task("Copy MIME files")
    def mime_files(log):
        mime_packages = Path(root, configuration.level_1_directory, "share", "mime", "packages")
        log.info(f"Using mime packages directory: {mime_packages}")
        os.makedirs(mime_packages, exist_ok=True)

        for file in Path(v.mime_xml_assets_dir()).glob("*.xml"):
            shutil.copyfile(str(file.absolute()), mime_packages.joinpath(file.name))

    @build.task("Copy icons")
    def copy_icons(log):
        icons = Path(root, configuration.level_1_directory, "share", "icons")
        log.info(f"Using icons directory: {icons}")

        shutil.copytree(v.icons_assets_dir(), icons)

    @build.task("Copy desktop entries")
    def copy_desktop_files(log):
        xdg_applications = Path(root, configuration.level_1_directory, "share", "applications")
        log.info(f"Using XDG applications directory: {xdg_applications}")
        os.makedirs(xdg_applications, exist_ok=True)

        desktop_variables = {
            "GRAPEJUICE_ICON": "grapejuice",
            "GRAPEJUICE_EXECUTABLE": os.path.join(
                configuration.target_system_root,
                *bin_path_components,
                grapejuice_exec_name
            ),
            "PLAYER_ICON": "grapejuice-roblox-player",
            "STUDIO_ICON": "grapejuice-roblox-studio"
        }

        for file in Path(v.desktop_assets_dir()).glob("*.desktop"):
            with file.open("r") as fp:
                template = Template(fp.read())
                finished_desktop_entry = template.substitute(desktop_variables)

            target_path = xdg_applications.joinpath(file.name)
            with target_path.open("w+") as fp:
                fp.write(finished_desktop_entry)

            target_path.chmod(0o755)

    @build.task("Copy binary entries")
    def copy_bin_scripts(log):
        usr_bin = Path(root, *bin_path_components)
        log.info(f"Using bin directory: {usr_bin}")
        os.makedirs(usr_bin, exist_ok=True)

        shutil.copyfile(res.bin_grapejuice_path(), usr_bin.joinpath(grapejuice_exec_name))
        shutil.copyfile(res.bin_grapejuiced_path(), usr_bin.joinpath("grapejuiced"))

        for file in usr_bin.glob("*"):
            file.chmod(0o755)

    build.run()


class LinuxPackageBuilder(PackageBuilder):
    def __init__(self, build_dir, dist_dir, configuration: LinuxPackageConfiguration = None):

        super().__init__(build_dir, dist_dir)

        self._configuration: LinuxPackageConfiguration = \
            configuration or LinuxPackageConfiguration(self._build_dir)

    def build(self):
        self.clean_build()
        self._prepare_build()

        _build_package(self._configuration)

    def dist(self):
        self.clean_dist()
        self._prepare_dist()
        path = Path(self._dist_dir, self._configuration.package_name)

        manifest = {
            "package": about.package_name,
            "package_version": about.package_version,
            "author": f"{about.author_name} <{about.author_email}>",
            "files": []
        }

        with tarfile.open(path, "w:gz") as tar:
            for file in Path(self._build_dir).rglob("*"):
                if file.is_dir():
                    continue

                file_path = str(file.absolute())
                if "__pycache__" in file_path:
                    continue

                arc_name = str(file.relative_to(self._build_dir))
                manifest["files"].append(arc_name)

                LOG.info(f"Adding to tar.gz: {file_path} -> {arc_name}")

                tar.add(
                    file_path,
                    arcname=arc_name
                )

            manifest_string = json.dumps(manifest).encode("UTF-8")
            manifest_info = tarfile.TarInfo(name="/".join([
                self._configuration.level_1_directory,
                "share",
                "grapejuice",
                "package_manifest.json"
            ]))
            manifest_info.size = len(manifest_string)

            buf = BytesIO()
            buf.write(manifest_string)
            buf.seek(0)

            tar.addfile(tarinfo=manifest_info, fileobj=buf)



from grapejuice_common import variables
from grapejuice_common.registry_file import RegistryFile


def logged_into_studio() -> bool:
    file = RegistryFile(variables.wine_user_reg())
    file.load()

    roblox_com = file.find_key(r"Software\\Roblox\\RobloxStudioBrowser\\roblox.com")
    if roblox_com:
        if roblox_com.get_attribute(".ROBLOSECURITY"):
            return True

    return False



from grapejuice import __version__ as grapejuice_version
from grapejuice_common import variables
from grapejuice_common.gtk.gtk_stuff import WindowBase


class AboutWindow(WindowBase):
    def __init__(self):
        super().__init__(variables.about_glade(), self)

        self.window.set_version(grapejuice_version)

    @property
    def window(self):
        return self.builder.get_object("grapejuice_about")

    def close_about(self, *_):
        self.window.destroy()

    def __del__(self):
        del self.builder



import json
import logging
import os
import shutil
import subprocess
import sys
from dataclasses import dataclass, asdict

from grapejuice_common import variables

LOG = logging.getLogger(__name__)


@dataclass
class UninstallationParameters:
    remove_prefix: bool
    for_reals: bool = False


def go(parameters: UninstallationParameters):
    assert parameters and isinstance(parameters, UninstallationParameters), "Programmer error: Invalid params argument"

    LOG.info("Uninstalling Grapejuice, parameters: " + json.dumps(asdict(parameters), indent=2))

    LOG.info(
        subprocess.check_output([
            sys.executable, "-m", "grapejuiced",
            "kill"
        ]).decode("UTF-8")
    )

    with open(variables.application_manifest(), "r") as fp:
        manifest = json.load(fp)

    for file in manifest["files"]:
        if not os.path.isabs(file):
            o_file = file
            file = os.path.join(variables.home(), file)
            LOG.info(f"Mended file path: {o_file} -> {file}")

        if os.path.exists(file) and os.path.isfile(file):
            LOG.info(f"Removing file from manifest: {file}")

            if parameters.for_reals:
                os.remove(file)

    if parameters.remove_prefix:
        LOG.info(f"Removing full user application directory: {variables.user_application_dir()}")

        if parameters.for_reals:
            shutil.rmtree(variables.user_application_dir(), ignore_errors=True)

    else:
        LOG.info(f"Removing manifest: {variables.application_manifest()}")

        if parameters.for_reals:
            os.remove(variables.application_manifest())

    LOG.info(
        subprocess.check_output([
            sys.executable, "-m", "pip",
            "uninstall", "-y", "grapejuice"
        ]).decode("UTF-8")
    )

    sys.exit(0)



import os
import subprocess
import sys

from grapejuice import background
from grapejuice_common import winectrl, variables
from grapejuice_common.update_info_providers import UpdateInformationProvider


def install_roblox():
    from grapejuice_common.ipc.dbus_client import dbus_connection
    dbus_connection().install_roblox()


class DisableMimeAssociations(background.BackgroundTask):
    def __init__(self, **kwargs):
        super().__init__("Disabling Wine associations", **kwargs)

    def work(self) -> None:
        winectrl.disable_mime_assoc()


class InstallRoblox(background.BackgroundTask):
    def __init__(self, **kwargs):
        super().__init__("Installing Roblox", **kwargs)

    def work(self) -> None:
        install_roblox()


class SandboxWine(background.BackgroundTask):
    def __init__(self, **kwargs):
        super().__init__("Sandboxing the Wine prefix", **kwargs)

    def work(self) -> None:
        winectrl.sandbox()


class RunRobloxStudio(background.BackgroundTask):
    def __init__(self, **kwargs):
        super().__init__("Launching Roblox Studio", **kwargs)

    def work(self) -> None:
        from grapejuice_common.ipc.dbus_client import dbus_connection
        dbus_connection().launch_studio()


class ExtractFastFlags(background.BackgroundTask):
    def __init__(self, **kwargs):
        super().__init__("Extracting Fast Flags", **kwargs)

    def work(self) -> None:
        from grapejuice_common.ipc.dbus_client import dbus_connection
        dbus_connection().extract_fast_flags()


class OpenLogsDirectory(background.BackgroundTask):
    def __init__(self, **kwargs):
        super().__init__("Opening logs directory", **kwargs)

    def work(self) -> None:
        path = variables.logging_directory()
        os.makedirs(path, exist_ok=True)

        subprocess.check_call(["xdg-open", path])


class PerformUpdate(background.BackgroundTask):
    def __init__(self, update_provider: UpdateInformationProvider, reopen: bool = False, **kwargs):
        super().__init__(**kwargs)
        self._update_provider = update_provider
        self._reopen = reopen

    def work(self) -> None:
        self._update_provider.do_update()

        if self._reopen:
            subprocess.Popen(["bash", "-c", f"{sys.executable} -m grapejuice gui & disown"], preexec_fn=os.setpgrp)

            from gi.repository import Gtk
            Gtk.main_quit()

            sys.exit(0)



import argparse
import logging
import os
import sys

from grapejuice_packaging.builders.appimage_builder import AppImageBuilder
from grapejuice_packaging.builders.debian_package_builder import DebianPackageBuilder
from grapejuice_packaging.builders.linux_package_builder import LinuxPackageBuilder
from grapejuice_packaging.builders.linux_supplemental_builder import LinuxSupplementalPackageBuilder
from grapejuice_packaging.builders.pypi_package_builder import PyPiPackageBuilder

logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)


def func_linux_package(args):
    build_dir = os.path.join(".", "build", "linux_package") if args.build_dir is None else args.build_dir
    dist_dir = os.path.join(".", "dist", "linux_package") if args.dist_dir is None else args.dist_dir

    builder = LinuxPackageBuilder(build_dir, dist_dir)

    builder.build()
    builder.dist()


def func_debian_package(args):
    build_dir = os.path.join(".", "build", "debian_package") if args.build_dir is None else args.build_dir
    dist_dir = os.path.join(".", "dist", "debian_package") if args.dist_dir is None else args.dist_dir

    builder = DebianPackageBuilder(build_dir, dist_dir)

    builder.build()
    builder.dist()


def func_supplemental_package(args):
    build_dir = os.path.join(".", "build", "supplemental_package") if args.build_dir is None else args.build_dir
    dist_dir = os.path.join(".", "dist", "supplemental_package") if args.dist_dir is None else args.dist_dir

    builder = LinuxSupplementalPackageBuilder(build_dir, dist_dir)

    builder.build()
    builder.dist()


def func_pypi_package(args):
    builder = PyPiPackageBuilder("build", "dist")
    builder.build()
    builder.dist()


def func_app_image(args):
    cwd = os.path.abspath(os.getcwd())
    build_dir = os.path.join(cwd, "build")
    dist_dir = os.path.join(cwd, "dist")

    builder = AppImageBuilder(build_dir, dist_dir)
    builder.build()
    builder.dist()


def main(in_args=None):
    if in_args is None:
        in_args = sys.argv

    parser = argparse.ArgumentParser(prog="grapejuice", description="Manage Roblox on Linux")
    subparsers = parser.add_subparsers(title="subcommands", help="sub-command help")

    parser_linux_package = subparsers.add_parser("linux_package")
    parser_linux_package.add_argument("--build-dir", required=False)
    parser_linux_package.add_argument("--dist-dir", required=False)
    parser_linux_package.set_defaults(func=func_linux_package)

    parser_debian_package = subparsers.add_parser("debian_package")
    parser_debian_package.add_argument("--build-dir", required=False)
    parser_debian_package.add_argument("--dist-dir", required=False)
    parser_debian_package.set_defaults(func=func_debian_package)

    parser_supplemental_package = subparsers.add_parser("supplemental_package")
    parser_supplemental_package.add_argument("--build-dir", required=False)
    parser_supplemental_package.add_argument("--dist-dir", required=False)
    parser_supplemental_package.set_defaults(func=func_supplemental_package)

    parser_app_image = subparsers.add_parser("app_image")
    parser_app_image.set_defaults(func=func_app_image)

    parser_pypi_package = subparsers.add_parser("pypi_package")
    parser_pypi_package.set_defaults(func=func_pypi_package)

    args = parser.parse_args(in_args[1:])

    if hasattr(args, "func"):
        f: callable = getattr(args, "func")
        return f(args) or 0

    else:
        parser.print_help()

    return 1


if __name__ == '__main__':
    sys.exit(main())



import logging


def log_function(func: callable):
    log = logging.getLogger(f"log_function/{func.__name__}")

    def wrapper(*args, **kwargs):
        log.debug(f"Calling function :: {repr(args)} :: {repr(kwargs)}")

        result = func(*args, **kwargs)

        if isinstance(result, tuple):
            result_list = list(result)

        else:
            result_list = [result]

        log.debug(f"Function result :: {repr(result_list)}")

        return result

    return wrapper


def log_on_call(message: str):
    def wrap_function(func: callable):
        log = logging.getLogger(f"log_on_call/{func.__name__}")

        def wrapper(*args, **kwargs):
            log.info(message)
            return func(*args, **kwargs)

        return wrapper

    return wrap_function



import sys

import grapejuice_common.variables as v
from grapejuice_packaging.builders.linux_package_builder import LinuxPackageBuilder, LinuxPackageConfiguration


class LinuxSupplementalPackageBuilder(LinuxPackageBuilder):
    def __init__(self, build_dir, dist_dir, level_1_directory: str = ".local"):
        configuration = LinuxPackageConfiguration(build_dir)
        configuration.python_site_type = "site-packages"
        configuration.python_site_version = f"python{sys.version_info.major}.{sys.version_info.minor}"
        configuration.copy_packages = False
        configuration.level_1_directory = level_1_directory
        configuration.target_system_root = v.home()

        super().__init__(build_dir, dist_dir, configuration)

    def dist(self):
        self.clean_dist()
        self._prepare_dist()

        super().dist()



def yes_no_dialog(title: str = "Untitled Dialog", message="This is a message"):
    from gi.repository import Gtk

    # Prepare message, it can be no more than 80 columns wide
    idx = 0
    line_limit = 80
    lines = [""]
    words = message.split(" ")

    while len(words) > 0:
        word = words.pop(0)

        if len(lines[idx]) > line_limit:
            lines[idx] = lines[idx]
            idx += 1
            lines.append(word + " ")

        else:
            lines[idx] += word + " "

    message = "\n".join(list(map(str.rstrip, lines)))

    # Create dialog class, duh
    class DialogClass(Gtk.Dialog):
        def __init__(self):
            Gtk.Dialog.__init__(self, title=title, flags=0)

            self.add_buttons(
                Gtk.STOCK_NO, Gtk.ResponseType.CANCEL,
                Gtk.STOCK_YES, Gtk.ResponseType.OK
            )
            self.set_resizable(False)

            self.set_default_size(150, 100)

            label = Gtk.Label(label=message)
            label.set_margin_top(10)
            label.set_margin_right(10)
            label.set_margin_bottom(10)
            label.set_margin_left(10)

            box = self.get_content_area()
            box.add(label)

            self.show_all()

    # Run the dialog
    dlg = DialogClass()
    response = dlg.run()
    dlg.destroy()

    # Process the response
    if response == Gtk.ResponseType.OK:
        return True

    elif response == Gtk.ResponseType.CANCEL:
        return False

    else:
        # WTF?
        raise RuntimeError("The universe is falling apart")



import os
import shutil
from abc import ABC, abstractmethod


class PackageBuilder(ABC):
    def __init__(self, build_dir, dist_dir):
        self._build_dir = os.path.abspath(build_dir)
        self._dist_dir = os.path.abspath(dist_dir)

    def _prepare_build(self):
        os.makedirs(self._build_dir, exist_ok=True)

    def _prepare_dist(self):
        os.makedirs(self._dist_dir, exist_ok=True)

    def clean_build(self):
        shutil.rmtree(self._build_dir, ignore_errors=True)

    def clean_dist(self):
        shutil.rmtree(self._dist_dir, ignore_errors=True)

    @abstractmethod
    def build(self):
        pass

    @abstractmethod
    def dist(self):
        pass



import os

HERE = os.path.dirname(os.path.abspath(__file__))


def bin_grapejuice_path():
    return os.path.join(HERE, "bin", "grapejuice")


def bin_grapejuiced_path():
    return os.path.join(HERE, "bin", "grapejuiced")


def app_image_app_run():
    return os.path.join(HERE, "appimage", "AppRun")


def app_image_desktop():
    return os.path.join(HERE, "appimage", "grapejuice.desktop")


def app_image_tool():  # Gross
    return os.path.join(HERE, "appimage", "appimagetool-x86_64.AppImage")



import logging
from abc import ABC, abstractmethod

LOG = logging.getLogger(__name__)


class IDBusConnection(ABC):
    @property
    def connected(self):
        return False

    @abstractmethod
    def launch_studio(self):
        pass

    @abstractmethod
    def play_game(self, uri):
        pass

    @abstractmethod
    def edit_local_game(self, place_path):
        pass

    @abstractmethod
    def edit_cloud_game(self, uri):
        pass

    @abstractmethod
    def install_roblox(self):
        pass

    @abstractmethod
    def version(self):
        pass

    @abstractmethod
    def extract_fast_flags(self):
        pass



import logging
import os
import sys

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
project_path = here
src_path = os.path.join(project_path, "src")
readme_path = os.path.join(project_path, "README.md")

sys.path.insert(0, src_path)

logging.basicConfig(stream=sys.stderr, level=logging.INFO)


def read_file(path):
    with open(path, "r") as fp:
        return fp.read()


def install_requires():
    requirements = read_file(os.path.join(project_path, "requirements.txt")).split("\n")
    requirements = list(filter(lambda s: not not s, map(lambda s: s.strip(), requirements)))

    return requirements


def main():
    import grapejuice.__about__ as __about__
    from grapejuice_packaging.local_install import InstallLocally

    setup(
        name="grapejuice",
        author=__about__.author_name,
        author_email=__about__.author_email,
        version=__about__.package_version,
        description=__about__.package_description,
        license=__about__.package_license,
        long_description=read_file(readme_path),
        long_description_content_type="text/markdown",
        url=__about__.package_repository,
        classifiers=[
            "Development Status :: 4 - Beta",
            'Programming Language :: Python :: 3.7'
        ],
        keywords=["grapejuice wine roblox studio"],
        packages=find_packages("src", exclude=[
            "grapejuice_packaging",
            "grapejuice_packaging.*",
            "tests",
            "tests.*"
        ]),
        package_dir={"": "src"},
        include_package_data=True,
        python_requires=">=3.7",
        install_requires=install_requires(),
        entry_points={
            "console_scripts": [
                "grapejuice=grapejuice.__main__:main",
                "grapejuiced=grapejuiced.__main__:main"
            ]
        },
        cmdclass={
            "install_locally": InstallLocally
        }
    )


if __name__ == '__main__':
    main()



from grapejuice_common.features import settings

connection = None


def dbus_connection():
    global connection

    if connection is None:
        from grapejuice_common.features.settings import current_settings

        if current_settings.get(settings.k_no_daemon_mode, True):
            from grapejuice_common.ipc.no_daemon_connection import NoDaemonModeConnection
            connection = NoDaemonModeConnection()

        else:
            from grapejuice_common.ipc.dbus_connection import DBusConnection
            connection = DBusConnection()

    return connection



import logging
import re
from pathlib import Path
from typing import Union, IO, List, Callable, Dict

LOG = logging.getLogger(__name__)

KEY_PTN = re.compile(r"\[(.*)?].*")


def line_iterator(fp: IO):
    for line in fp.read().split("\n"):
        yield line.strip("\r").strip()


class RegistryKey:
    _path: str
    _value: any
    _attributes: Dict[str, str]

    def __init__(self, path: str):
        self._path = path
        self._attributes = dict()

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v):
        self._value = v

    def set_attribute(self, key: str, value: str):
        self._attributes[key] = value

    def get_attribute(self, key: str):
        return self._attributes.get(key, None)


class RegistryFile:
    _path: Path

    _version: str = ""
    _comments: List[str]
    _root_key: RegistryKey
    _keys: Dict[str, RegistryKey]
    _last_key_read: Union[RegistryKey, None] = None

    def __init__(self, path: Union[str, Path]):
        if isinstance(path, str):
            self._path = Path(path).absolute()

        else:
            self._path = path.absolute()

        self._comments = []
        self._keys = dict()
        self._root_key = RegistryKey("\\")

    @property
    def _current_key(self) -> RegistryKey:
        if self._last_key_read is not None:
            return self._last_key_read

        else:
            return self._root_key

    def find_key(self, path: str):
        return self._keys.get(path, None)

    def load(self):
        class LoadState:
            operator: Callable[[str], None]

        def normal(ln: str):
            if ln.startswith(";;"):
                return comment(ln)

            match = KEY_PTN.match(ln)
            if match:
                return key(ln)

            return attribute(ln)

        def version(ln: str):
            self._version = ln
            LoadState.operator = normal

        def comment(ln: str):
            self._comments.append(ln)

        def key(ln: str):
            split = ln.split(" ")
            path = split[0].lstrip("[").strip("]")
            registry_key = self._keys.setdefault(path, RegistryKey(path))

            if len(split) > 1:
                registry_key.value = split[1].strip()

            self._last_key_read = registry_key

        def attribute(ln: str):
            if not ln:
                return

            match = re.match(r"#(.*)?=(.*)", ln)
            if match:
                self._current_key.set_attribute(match.group(1), match.group(2))
                return

            match = re.match(r"\"(.*)?\"\s*=\s*(.*)", ln)
            if match:
                self._current_key.set_attribute(match.group(1), match.group(2))
                return

        LoadState.operator = version

        with self._path.open("r") as fp:
            if callable(LoadState.operator):
                for line in line_iterator(fp):
                    LoadState.operator(line)

            else:
                LOG.warning("LoadState.operator is not callable")



import io
import logging
import os
import re
import shutil
import subprocess
import sys
import tarfile
from abc import ABC, abstractmethod
from pathlib import Path

import requests
from packaging import version

from grapejuice_common import variables
from grapejuice_common.features import settings
from grapejuice_common.logs.log_util import log_function

LOG = logging.getLogger(__name__)

VERSION_PTN = re.compile(r"__version__\s*=\s*\"([\d\.]+)\".*")


class UpdateError(RuntimeError):
    pass


class UpdateInformationProvider(ABC):
    _cached_gitlab_version: version.Version = None

    @staticmethod
    def local_version() -> version.Version:
        from grapejuice.__about__ import gj_package_version
        return version.parse(gj_package_version)

    @staticmethod
    def gitlab_version(return_cached: bool = False) -> version.Version:
        if return_cached and UpdateInformationProvider._cached_gitlab_version is not None:
            return UpdateInformationProvider._cached_gitlab_version

        url = variables.git_grapejuice_init()
        response = requests.get(url)

        if response.status_code < 200 or response.status_code > 299:
            LOG.error(
                "Failed to get the version of grapejuice on GitLab. Returning version 0\n"
                f"URL: {url}\n"
                f"Response text: {response.text}\n"
            )

            return version.parse("0.0.0")

        for line in response.text.replace("\r", "").split("\n"):
            match = VERSION_PTN.match(line)
            if not match:
                continue

            ver = version.parse(match.group(1).strip())
            UpdateInformationProvider._cached_gitlab_version = ver

            return ver

        LOG.error("Could not match a Grapejuice version string in the remote repository")
        LOG.warning("Returning version 0.0.0")

        return version.parse("0.0.0")

    @staticmethod
    def can_update() -> bool:
        return False

    @abstractmethod
    def do_update(self) -> None:
        pass

    @abstractmethod
    def update_available(self) -> bool:
        pass

    @abstractmethod
    def local_is_newer(self) -> bool:
        pass

    @abstractmethod
    def target_version(self) -> version.Version:
        pass


class SourceUpdateInformationProvider(UpdateInformationProvider):
    def target_version(self) -> version.Version:
        return UpdateInformationProvider.gitlab_version(return_cached=True)

    def update_available(self) -> bool:
        return UpdateInformationProvider.gitlab_version() > UpdateInformationProvider.local_version()

    def local_is_newer(self) -> bool:
        return UpdateInformationProvider.local_version() > UpdateInformationProvider.gitlab_version(return_cached=True)

    @staticmethod
    def can_update() -> bool:
        return True

    def do_update(self):
        from grapejuice_common.features.settings import current_settings

        tmp_path = variables.tmp_path()
        LOG.info(f"Temporary files path at: {tmp_path}")
        update_package_path = os.path.join(tmp_path, "update")

        response = requests.get(variables.git_source_tarball())
        if response.status_code < 200 or response.status_code > 299:
            raise UpdateError(f"Received HTTP error {response.status_code} from GitLab")

        if os.path.exists(update_package_path):
            LOG.warning(f"Removing existing update package: {update_package_path}")
            shutil.rmtree(update_package_path, ignore_errors=True)

        else:
            LOG.debug(f"Creating update package directory: {update_package_path}")
            os.makedirs(update_package_path)

        fp = io.BytesIO(response.content)
        with tarfile.open(fileobj=fp) as tar:
            tar.extractall(update_package_path)

        cwd = os.getcwd()

        release_channel = current_settings.get(settings.k_release_channel)
        os.chdir(os.path.join(update_package_path, f"grapejuice-{release_channel}"))

        LOG.debug("Installing update")
        subprocess.check_call([sys.executable, "./install.py"])
        os.chdir(cwd)

        fp.close()
        del fp
        del response

        shutil.rmtree(tmp_path)


class NonUpgradablePackageInformationProvider(UpdateInformationProvider):
    def target_version(self) -> version.Version:
        return version.parse("0.0.0")

    def update_available(self) -> bool:
        return False

    def local_is_newer(self) -> bool:
        return False

    @staticmethod
    def can_update() -> bool:
        return False

    def do_update(self) -> None:
        raise UpdateError("Grapejuice cannot upgrade a system package")


class SystemUpdateInformationProvider(NonUpgradablePackageInformationProvider):
    pass


@log_function
def guess_relevant_provider() -> UpdateInformationProvider:
    this_file = Path(__file__).resolve()
    stat = this_file.stat()

    if stat.st_uid == 0 and stat.st_gid == 0:
        return SystemUpdateInformationProvider()

    elif str(this_file).startswith(variables.home()):
        return SourceUpdateInformationProvider()

    else:
        return NonUpgradablePackageInformationProvider()



import argparse
import signal
import subprocess
import sys

from grapejuice_common.ipc.pid_file import PIDFile, daemon_pid_file
from grapejuice_common.logs import self_test, log_config


def spawn(pid_file: PIDFile):
    from grapejuiced.state import State
    state = State()

    def on_sigint(signum, frame) -> None:
        print("> Responding to SIGINT, stopping...")
        state.stop()

    signal.signal(signal.SIGINT, on_sigint)

    print("> Spawning a new daemon")
    pid_file.write_pid()
    state.start_service()
    state.start()


def func_kill(*_):
    print("> You swing your sword...")

    pid_file = daemon_pid_file()
    if pid_file.is_running():
        print("> Killed daemon with pid {} in one sweeping blow.".format(pid_file.pid))
        pid_file.kill()

    else:
        print("> You swing at the air, because there is no daemon. You take 20 damage as you hit your leg.")


def func_daemon(args):
    if args.kill:
        subprocess.check_call([
            sys.executable, "-m", "grapejuiced", "kill"
        ])

    pid_file = daemon_pid_file()

    if pid_file.is_running():
        print("> Another daemon is already running, quitting...")
        return

    self_test.post.run()

    spawn(pid_file)


def main(in_args=None):
    log_config.configure_logging("grapejuice-daemon")

    if in_args is None:
        in_args = sys.argv

    parser = argparse.ArgumentParser(prog="grapejuiced", description="The Grapejuice daemon")
    subparsers = parser.add_subparsers(title="subcommands", help="sub-command help")

    parser_kill = subparsers.add_parser("kill")
    parser_kill.set_defaults(func=func_kill)

    parser_daemon = subparsers.add_parser("daemonize")
    parser_daemon.add_argument("--kill", action="store_true")
    parser_daemon.set_defaults(func=func_daemon)

    args = parser.parse_args(in_args[1:])
    if hasattr(args, "func"):
        f: callable = getattr(args, "func")
        f(args)

    else:
        parser.print_help()


if __name__ == '__main__':
    main()
